@@ -10,8 +10,6 @@ use num::{NumCast, ToPrimitive};
 use std::ops::Not;
 use std::sync::Arc;
 
-type LargeStringArray = Utf8Array<i64>;
-
 impl<T> ChunkedArray<T>
 where
     T: PolarsNumericType,
@@ -32,7 +30,7 @@ where
             })
             .collect::<Vec<_>>();
 
-        ChunkedArray::new_from_chunks("", chunks)
+        ChunkedArray::from_chunks("", chunks)
     }
 }
 
@@ -81,7 +79,7 @@ where
             _ => {
                 // same length
                 let (lhs, rhs) = align_chunks_binary(self, rhs);
-                lhs.comparison(&rhs, |x, y| comparison::eq(x, y))
+                lhs.comparison(&rhs, |x, y| comparison::eq_and_validity(x, y))
             }
         }
     }
@@ -106,7 +104,7 @@ where
             _ => {
                 // same length
                 let (lhs, rhs) = align_chunks_binary(self, rhs);
-                lhs.comparison(&rhs, |x, y| comparison::neq(x, y))
+                lhs.comparison(&rhs, |x, y| comparison::neq_and_validity(x, y))
             }
         }
     }
@@ -223,7 +221,7 @@ fn compare_bools(
         .map(|(l, r)| Arc::new(f(l, r)) as ArrayRef)
         .collect();
 
-    BooleanChunked::new_from_chunks(lhs.name(), chunks)
+    BooleanChunked::from_chunks(lhs.name(), chunks)
 }
 
 impl ChunkCompare<&BooleanChunked> for BooleanChunked {
@@ -257,7 +255,7 @@ impl ChunkCompare<&BooleanChunked> for BooleanChunked {
             _ => {
                 // same length
                 let (lhs, rhs) = align_chunks_binary(self, rhs);
-                compare_bools(&lhs, &rhs, |lhs, rhs| comparison::eq(lhs, rhs))
+                compare_bools(&lhs, &rhs, |lhs, rhs| comparison::eq_and_validity(lhs, rhs))
             }
         }
     }
@@ -288,7 +286,9 @@ impl ChunkCompare<&BooleanChunked> for BooleanChunked {
             _ => {
                 // same length
                 let (lhs, rhs) = align_chunks_binary(self, rhs);
-                compare_bools(&lhs, &rhs, |lhs, rhs| comparison::neq(lhs, rhs))
+                compare_bools(&lhs, &rhs, |lhs, rhs| {
+                    comparison::neq_and_validity(lhs, rhs)
+                })
             }
         }
     }
@@ -425,24 +425,14 @@ impl Utf8Chunked {
         f: impl Fn(&Utf8Array<i64>, &Utf8Array<i64>) -> BooleanArray,
     ) -> BooleanChunked {
         let chunks = self
-            .chunks
-            .iter()
-            .zip(&rhs.chunks)
+            .downcast_iter()
+            .zip(rhs.downcast_iter())
             .map(|(left, right)| {
-                let left = left
-                    .as_any()
-                    .downcast_ref::<LargeStringArray>()
-                    .expect("could not downcast one of the chunks");
-                let right = right
-                    .as_any()
-                    .downcast_ref::<LargeStringArray>()
-                    .expect("could not downcast one of the chunks");
                 let arr = f(left, right);
                 Arc::new(arr) as ArrayRef
             })
-            .collect::<Vec<_>>();
-
-        ChunkedArray::new_from_chunks("", chunks)
+            .collect();
+        BooleanChunked::from_chunks("", chunks)
     }
 }
 
@@ -465,12 +455,9 @@ impl ChunkCompare<&Utf8Chunked> for Utf8Chunked {
             } else {
                 BooleanChunked::full("", false, self.len())
             }
-        }
-        // same length
-        else if self.chunk_id().zip(rhs.chunk_id()).all(|(l, r)| l == r) {
-            self.comparison(rhs, |l, r| comparison::eq(l, r))
         } else {
-            apply_operand_on_chunkedarray_by_iter!(self, rhs, ==)
+            let (lhs, rhs) = align_chunks_binary(self, rhs);
+            lhs.comparison(&rhs, comparison::utf8::eq_and_validity)
         }
     }
 
@@ -488,12 +475,9 @@ impl ChunkCompare<&Utf8Chunked> for Utf8Chunked {
             } else {
                 BooleanChunked::full("", false, self.len())
             }
-        }
-        // same length
-        else if self.chunk_id().zip(rhs.chunk_id()).all(|(l, r)| l == r) {
-            self.comparison(rhs, |l, r| comparison::neq(l, r))
         } else {
-            apply_operand_on_chunkedarray_by_iter!(self, rhs, !=)
+            let (lhs, rhs) = align_chunks_binary(self, rhs);
+            lhs.comparison(&rhs, comparison::utf8::neq_and_validity)
         }
     }
 
@@ -602,7 +586,7 @@ where
         let rhs: T::Native =
             NumCast::from(rhs).expect("could not cast to underlying chunkedarray type");
         let scalar = PrimitiveScalar::new(T::get_dtype().to_arrow(), Some(rhs));
-        self.apply_kernel_cast(|arr| Arc::new(f(arr, &scalar)))
+        self.apply_kernel_cast(&|arr| Arc::new(f(arr, &scalar)))
     }
 }
 
@@ -616,11 +600,11 @@ where
     }
 
     fn equal(&self, rhs: Rhs) -> BooleanChunked {
-        self.primitive_compare_scalar(rhs, |l, rhs| comparison::eq_scalar(l, rhs))
+        self.primitive_compare_scalar(rhs, |l, rhs| comparison::eq_scalar_and_validity(l, rhs))
     }
 
     fn not_equal(&self, rhs: Rhs) -> BooleanChunked {
-        self.primitive_compare_scalar(rhs, |l, rhs| comparison::neq_scalar(l, rhs))
+        self.primitive_compare_scalar(rhs, |l, rhs| comparison::neq_scalar_and_validity(l, rhs))
     }
 
     fn gt(&self, rhs: Rhs) -> BooleanChunked {
@@ -647,7 +631,7 @@ impl Utf8Chunked {
         f: impl Fn(&Utf8Array<i64>, &dyn Scalar) -> BooleanArray,
     ) -> BooleanChunked {
         let scalar = Utf8Scalar::<i64>::new(Some(rhs));
-        self.apply_kernel_cast(|arr| Arc::new(f(arr, &scalar)))
+        self.apply_kernel_cast(&|arr| Arc::new(f(arr, &scalar)))
     }
 }
 
@@ -657,10 +641,10 @@ impl ChunkCompare<&str> for Utf8Chunked {
     }
 
     fn equal(&self, rhs: &str) -> BooleanChunked {
-        self.utf8_compare_scalar(rhs, |l, rhs| comparison::eq_scalar(l, rhs))
+        self.utf8_compare_scalar(rhs, |l, rhs| comparison::eq_scalar_and_validity(l, rhs))
     }
     fn not_equal(&self, rhs: &str) -> BooleanChunked {
-        self.utf8_compare_scalar(rhs, |l, rhs| comparison::neq_scalar(l, rhs))
+        self.utf8_compare_scalar(rhs, |l, rhs| comparison::neq_scalar_and_validity(l, rhs))
     }
 
     fn gt(&self, rhs: &str) -> BooleanChunked {
@@ -780,7 +764,7 @@ impl Not for &BooleanChunked {
                 Arc::new(arr) as ArrayRef
             })
             .collect::<Vec<_>>();
-        ChunkedArray::new_from_chunks(self.name(), chunks)
+        ChunkedArray::from_chunks(self.name(), chunks)
     }
 }
 
@@ -806,6 +790,7 @@ impl BooleanChunked {
 
 // private
 pub(crate) trait ChunkEqualElement {
+    /// Only meant for physical types.
     /// Check if element in self is equal to element in other, assumes same dtypes
     ///
     /// # Safety
@@ -848,15 +833,6 @@ impl ChunkEqualElement for Utf8Chunked {
 }
 
 impl ChunkEqualElement for ListChunked {}
-#[cfg(feature = "dtype-categorical")]
-impl ChunkEqualElement for CategoricalChunked {
-    unsafe fn equal_element(&self, idx_self: usize, idx_other: usize, other: &Series) -> bool {
-        let ca_other = other.as_ref().as_ref();
-        debug_assert!(self.dtype() == other.dtype());
-        let ca_other = &*(ca_other as *const CategoricalChunked);
-        self.get(idx_self) == ca_other.get(idx_other)
-    }
-}
 
 #[cfg(test)]
 mod test {
@@ -1074,8 +1050,8 @@ mod test {
     #[test]
     fn test_kleene() {
         let a = BooleanChunked::new("", &[Some(true), Some(false), None]);
-        let trues = BooleanChunked::new_from_slice("", &[true, true, true]);
-        let falses = BooleanChunked::new_from_slice("", &[false, false, false]);
+        let trues = BooleanChunked::from_slice("", &[true, true, true]);
+        let falses = BooleanChunked::from_slice("", &[false, false, false]);
 
         let c = &a | &trues;
         assert_eq!(Vec::from(&c), &[Some(true), Some(true), Some(true)]);
@@ -1086,9 +1062,9 @@ mod test {
 
     #[test]
     fn test_broadcasting_bools() {
-        let a = BooleanChunked::new_from_slice("", &[true, false, true]);
-        let true_ = BooleanChunked::new_from_slice("", &[true]);
-        let false_ = BooleanChunked::new_from_slice("", &[false]);
+        let a = BooleanChunked::from_slice("", &[true, false, true]);
+        let true_ = BooleanChunked::from_slice("", &[true]);
+        let false_ = BooleanChunked::from_slice("", &[false]);
 
         let out = a.equal(&true_);
         assert_eq!(Vec::from(&out), &[Some(true), Some(false), Some(true)]);
@@ -1147,9 +1123,9 @@ mod test {
 
     #[test]
     fn test_broadcasting_numeric() {
-        let a = Int32Chunked::new_from_slice("", &[1, 2, 3]);
-        let one = Int32Chunked::new_from_slice("", &[1]);
-        let three = Int32Chunked::new_from_slice("", &[3]);
+        let a = Int32Chunked::from_slice("", &[1, 2, 3]);
+        let one = Int32Chunked::from_slice("", &[1]);
+        let three = Int32Chunked::from_slice("", &[3]);
 
         let out = a.equal(&one);
         assert_eq!(Vec::from(&out), &[Some(true), Some(false), Some(false)]);
