@@ -1,11 +1,7 @@
-#[cfg(feature = "dtype-categorical")]
-use crate::datatypes::CategoricalChunked;
 use crate::prelude::*;
 use crate::utils::CustomIterTools;
 use arrow::array::*;
 use std::convert::TryFrom;
-#[cfg(feature = "dtype-categorical")]
-use std::ops::Deref;
 
 type LargeStringArray = Utf8Array<i64>;
 type LargeListArray = ListArray<i64>;
@@ -37,29 +33,23 @@ where
     type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
     fn into_iter(self) -> Self::IntoIter {
         Box::new(
-            self.downcast_iter()
-                .flatten()
-                .map(|x| x.copied())
-                .trust_my_length(self.len()),
+            // we know that we only iterate over length == self.len()
+            unsafe {
+                self.downcast_iter()
+                    .flatten()
+                    .map(|x| x.copied())
+                    .trust_my_length(self.len())
+            },
         )
     }
 }
 
-#[cfg(feature = "dtype-categorical")]
-impl<'a> IntoIterator for &'a CategoricalChunked {
-    type Item = Option<u32>;
-    type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
-
-    fn into_iter(self) -> Self::IntoIter {
-        self.deref().into_iter()
-    }
-}
-
 impl<'a> IntoIterator for &'a BooleanChunked {
     type Item = Option<bool>;
     type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
     fn into_iter(self) -> Self::IntoIter {
-        Box::new(self.downcast_iter().flatten().trust_my_length(self.len()))
+        // we know that we only iterate over length == self.len()
+        unsafe { Box::new(self.downcast_iter().flatten().trust_my_length(self.len())) }
     }
 }
 
@@ -118,6 +108,7 @@ impl<'a> ExactSizeIterator for BoolIterNoNull<'a> {}
 
 impl BooleanChunked {
     #[allow(clippy::wrong_self_convention)]
+    #[doc(hidden)]
     pub fn into_no_null_iter(
         &self,
     ) -> impl Iterator<Item = bool>
@@ -127,10 +118,12 @@ impl BooleanChunked {
            + ExactSizeIterator
            + DoubleEndedIterator
            + TrustedLen {
-        self.downcast_iter()
-            .map(BoolIterNoNull::new)
-            .flatten()
-            .trust_my_length(self.len())
+        // we know that we only iterate over length == self.len()
+        unsafe {
+            self.downcast_iter()
+                .flat_map(BoolIterNoNull::new)
+                .trust_my_length(self.len())
+        }
     }
 }
 
@@ -138,7 +131,8 @@ impl<'a> IntoIterator for &'a Utf8Chunked {
     type Item = Option<&'a str>;
     type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
     fn into_iter(self) -> Self::IntoIter {
-        Box::new(self.downcast_iter().flatten().trust_my_length(self.len()))
+        // we know that we only iterate over length == self.len()
+        unsafe { Box::new(self.downcast_iter().flatten().trust_my_length(self.len())) }
     }
 }
 
@@ -196,6 +190,7 @@ impl<'a> ExactSizeIterator for Utf8IterNoNull<'a> {}
 
 impl Utf8Chunked {
     #[allow(clippy::wrong_self_convention)]
+    #[doc(hidden)]
     pub fn into_no_null_iter<'a>(
         &'a self,
     ) -> impl Iterator<Item = &'a str>
@@ -205,10 +200,12 @@ impl Utf8Chunked {
            + ExactSizeIterator
            + DoubleEndedIterator
            + TrustedLen {
-        self.downcast_iter()
-            .map(Utf8IterNoNull::new)
-            .flatten()
-            .trust_my_length(self.len())
+        // we know that we only iterate over length == self.len()
+        unsafe {
+            self.downcast_iter()
+                .flat_map(Utf8IterNoNull::new)
+                .trust_my_length(self.len())
+        }
     }
 }
 
@@ -216,27 +213,41 @@ impl<'a> IntoIterator for &'a ListChunked {
     type Item = Option<Series>;
     type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
     fn into_iter(self) -> Self::IntoIter {
-        Box::new(
-            self.downcast_iter()
-                .map(|arr| arr.iter())
-                .flatten()
-                .trust_my_length(self.len())
-                .map(|arr| arr.map(|arr| Series::try_from(("", arr)).unwrap())),
-        )
+        let dtype = self.inner_dtype();
+
+        // we know that we only iterate over length == self.len()
+        unsafe {
+            Box::new(
+                self.downcast_iter()
+                    .flat_map(|arr| arr.iter())
+                    .trust_my_length(self.len())
+                    .map(move |arr| {
+                        arr.map(|arr| {
+                            Series::from_chunks_and_dtype_unchecked(
+                                "",
+                                vec![Arc::from(arr)],
+                                &dtype,
+                            )
+                        })
+                    }),
+            )
+        }
     }
 }
 
 pub struct ListIterNoNull<'a> {
     array: &'a LargeListArray,
+    inner_type: DataType,
     current: usize,
     current_end: usize,
 }
 
 impl<'a> ListIterNoNull<'a> {
     /// create a new iterator
-    pub fn new(array: &'a LargeListArray) -> Self {
+    pub fn new(array: &'a LargeListArray, inner_type: DataType) -> Self {
         ListIterNoNull {
             array,
+            inner_type,
             current: 0,
             current_end: array.len(),
         }
@@ -252,7 +263,13 @@ impl<'a> Iterator for ListIterNoNull<'a> {
         } else {
             let old = self.current;
             self.current += 1;
-            unsafe { Some(Series::try_from(("", self.array.value_unchecked(old))).unwrap()) }
+            unsafe {
+                Some(Series::from_chunks_and_dtype_unchecked(
+                    "",
+                    vec![Arc::from(self.array.value_unchecked(old))],
+                    &self.inner_type,
+                ))
+            }
         }
     }
 
@@ -282,6 +299,7 @@ impl<'a> ExactSizeIterator for ListIterNoNull<'a> {}
 
 impl ListChunked {
     #[allow(clippy::wrong_self_convention)]
+    #[doc(hidden)]
     pub fn into_no_null_iter(
         &self,
     ) -> impl Iterator<Item = Series>
@@ -291,10 +309,13 @@ impl ListChunked {
            + ExactSizeIterator
            + DoubleEndedIterator
            + TrustedLen {
-        self.downcast_iter()
-            .map(ListIterNoNull::new)
-            .flatten()
-            .trust_my_length(self.len())
+        // we know that we only iterate over length == self.len()
+        let inner_type = self.inner_dtype();
+        unsafe {
+            self.downcast_iter()
+                .flat_map(move |arr| ListIterNoNull::new(arr, inner_type.clone()))
+                .trust_my_length(self.len())
+        }
     }
 }
 
@@ -306,33 +327,28 @@ where
     type Item = Option<&'a T>;
     type IntoIter = Box<dyn PolarsIterator<Item = Self::Item> + 'a>;
     fn into_iter(self) -> Self::IntoIter {
-        Box::new(self.downcast_iter().flatten().trust_my_length(self.len()))
+        // we know that we only iterate over length == self.len()
+        unsafe { Box::new(self.downcast_iter().flatten().trust_my_length(self.len())) }
     }
 }
 
 #[cfg(feature = "object")]
 impl<T: PolarsObject> ObjectChunked<T> {
     #[allow(clippy::wrong_self_convention)]
+    #[doc(hidden)]
     pub fn into_no_null_iter(
         &self,
     ) -> impl Iterator<Item = &T> + '_ + Send + Sync + ExactSizeIterator + DoubleEndedIterator + TrustedLen
     {
-        self.downcast_iter()
-            .map(|arr| arr.values().iter())
-            .flatten()
-            .trust_my_length(self.len())
+        // we know that we only iterate over length == self.len()
+        unsafe {
+            self.downcast_iter()
+                .flat_map(|arr| arr.values().iter())
+                .trust_my_length(self.len())
+        }
     }
 }
 
-/// Trait for ChunkedArrays that don't have null values.
-/// The result is the most efficient implementation `Iterator`, according to the number of chunks.
-pub trait IntoNoNullIterator {
-    type Item;
-    type IntoIter: Iterator<Item = Self::Item>;
-
-    fn into_no_null_iter(self) -> Self::IntoIter;
-}
-
 /// Wrapper struct to convert an iterator of type `T` into one of type `Option<T>`.  It is useful to make the
 /// `IntoIterator` trait, in which every iterator shall return an `Option<T>`.
 pub struct SomeIterator<I>(I)
@@ -365,30 +381,14 @@ where
 
 impl<I> ExactSizeIterator for SomeIterator<I> where I: ExactSizeIterator {}
 
-#[cfg(feature = "dtype-categorical")]
-impl CategoricalChunked {
-    #[allow(clippy::wrong_self_convention)]
-    pub fn into_no_null_iter(
-        &self,
-    ) -> impl Iterator<Item = u32>
-           + '_
-           + Send
-           + Sync
-           + ExactSizeIterator
-           + DoubleEndedIterator
-           + TrustedLen {
-        self.deref().into_no_null_iter()
-    }
-}
-
 #[cfg(test)]
 mod test {
     use crate::prelude::*;
 
     #[test]
     fn out_of_bounds() {
-        let mut a = UInt32Chunked::new_from_slice("a", &[1, 2, 3]);
-        let b = UInt32Chunked::new_from_slice("a", &[1, 2, 3]);
+        let mut a = UInt32Chunked::from_slice("a", &[1, 2, 3]);
+        let b = UInt32Chunked::from_slice("a", &[1, 2, 3]);
         a.append(&b);
 
         let v = a.into_iter().collect::<Vec<_>>();
@@ -413,7 +413,7 @@ mod test {
         ($test_name:ident, $ca_type:ty, $first_val:expr, $second_val:expr, $third_val:expr) => {
             #[test]
             fn $test_name() {
-                let a = <$ca_type>::new_from_slice("test", &[$first_val, $second_val, $third_val]);
+                let a = <$ca_type>::from_slice("test", &[$first_val, $second_val, $third_val]);
 
                 // normal iterator
                 let mut it = a.into_iter();
@@ -474,8 +474,7 @@ mod test {
         ($test_name:ident, $ca_type:ty, $first_val:expr, $second_val:expr, $third_val:expr) => {
             #[test]
             fn $test_name() {
-                let a =
-                    <$ca_type>::new_from_opt_slice("test", &[$first_val, $second_val, $third_val]);
+                let a = <$ca_type>::new("test", &[$first_val, $second_val, $third_val]);
 
                 // normal iterator
                 let mut it = a.into_iter();
@@ -554,8 +553,8 @@ mod test {
         ($test_name:ident, $ca_type:ty, $first_val:expr, $second_val:expr, $third_val:expr) => {
             #[test]
             fn $test_name() {
-                let mut a = <$ca_type>::new_from_slice("test", &[$first_val, $second_val]);
-                let a_b = <$ca_type>::new_from_slice("", &[$third_val]);
+                let mut a = <$ca_type>::from_slice("test", &[$first_val, $second_val]);
+                let a_b = <$ca_type>::from_slice("", &[$third_val]);
                 a.append(&a_b);
 
                 // normal iterator
@@ -617,8 +616,8 @@ mod test {
         ($test_name:ident, $ca_type:ty, $first_val:expr, $second_val:expr, $third_val:expr) => {
             #[test]
             fn $test_name() {
-                let mut a = <$ca_type>::new_from_opt_slice("test", &[$first_val, $second_val]);
-                let a_b = <$ca_type>::new_from_opt_slice("", &[$third_val]);
+                let mut a = <$ca_type>::new("test", &[$first_val, $second_val]);
+                let a_b = <$ca_type>::new("", &[$third_val]);
                 a.append(&a_b);
 
                 // normal iterator
@@ -698,7 +697,7 @@ mod test {
         ($test_name:ident, $ca_type:ty, $first_val:expr, $second_val:expr, $third_val:expr) => {
             #[test]
             fn $test_name() {
-                let a = <$ca_type>::new_from_slice("test", &[$first_val, $second_val, $third_val]);
+                let a = <$ca_type>::from_slice("test", &[$first_val, $second_val, $third_val]);
 
                 // normal iterator
                 let mut it = a.into_no_null_iter();
@@ -771,8 +770,8 @@ mod test {
         ($test_name:ident, $ca_type:ty, $first_val:expr, $second_val:expr, $third_val:expr) => {
             #[test]
             fn $test_name() {
-                let mut a = <$ca_type>::new_from_slice("test", &[$first_val, $second_val]);
-                let a_b = <$ca_type>::new_from_slice("", &[$third_val]);
+                let mut a = <$ca_type>::from_slice("test", &[$first_val, $second_val]);
+                let a_b = <$ca_type>::from_slice("", &[$third_val]);
                 a.append(&a_b);
 
                 // normal iterator
@@ -878,7 +877,7 @@ mod test {
     }
 
     impl_test_iter_skip!(utf8_iter_single_chunk_skip, 8, Some("0"), Some("9"), {
-        Utf8Chunked::new_from_slice("test", &generate_utf8_vec(SKIP_ITERATOR_SIZE))
+        Utf8Chunked::from_slice("test", &generate_utf8_vec(SKIP_ITERATOR_SIZE))
     });
 
     impl_test_iter_skip!(
@@ -886,21 +885,19 @@ mod test {
         8,
         Some("0"),
         None,
-        { Utf8Chunked::new_from_opt_slice("test", &generate_opt_utf8_vec(SKIP_ITERATOR_SIZE)) }
+        { Utf8Chunked::new("test", &generate_opt_utf8_vec(SKIP_ITERATOR_SIZE)) }
     );
 
     impl_test_iter_skip!(utf8_iter_many_chunk_skip, 18, Some("0"), Some("9"), {
-        let mut a = Utf8Chunked::new_from_slice("test", &generate_utf8_vec(SKIP_ITERATOR_SIZE));
-        let a_b = Utf8Chunked::new_from_slice("test", &generate_utf8_vec(SKIP_ITERATOR_SIZE));
+        let mut a = Utf8Chunked::from_slice("test", &generate_utf8_vec(SKIP_ITERATOR_SIZE));
+        let a_b = Utf8Chunked::from_slice("test", &generate_utf8_vec(SKIP_ITERATOR_SIZE));
         a.append(&a_b);
         a
     });
 
     impl_test_iter_skip!(utf8_iter_many_chunk_null_check_skip, 18, Some("0"), None, {
-        let mut a =
-            Utf8Chunked::new_from_opt_slice("test", &generate_opt_utf8_vec(SKIP_ITERATOR_SIZE));
-        let a_b =
-            Utf8Chunked::new_from_opt_slice("test", &generate_opt_utf8_vec(SKIP_ITERATOR_SIZE));
+        let mut a = Utf8Chunked::new("test", &generate_opt_utf8_vec(SKIP_ITERATOR_SIZE));
+        let a_b = Utf8Chunked::new("test", &generate_opt_utf8_vec(SKIP_ITERATOR_SIZE));
         a.append(&a_b);
         a
     });
@@ -921,30 +918,23 @@ mod test {
     }
 
     impl_test_iter_skip!(bool_iter_single_chunk_skip, 8, Some(true), Some(false), {
-        BooleanChunked::new_from_slice("test", &generate_boolean_vec(SKIP_ITERATOR_SIZE))
+        BooleanChunked::from_slice("test", &generate_boolean_vec(SKIP_ITERATOR_SIZE))
     });
 
     impl_test_iter_skip!(bool_iter_single_chunk_null_check_skip, 8, None, None, {
-        BooleanChunked::new_from_opt_slice("test", &generate_opt_boolean_vec(SKIP_ITERATOR_SIZE))
+        BooleanChunked::new("test", &generate_opt_boolean_vec(SKIP_ITERATOR_SIZE))
     });
 
     impl_test_iter_skip!(bool_iter_many_chunk_skip, 18, Some(true), Some(false), {
-        let mut a =
-            BooleanChunked::new_from_slice("test", &generate_boolean_vec(SKIP_ITERATOR_SIZE));
-        let a_b = BooleanChunked::new_from_slice("test", &generate_boolean_vec(SKIP_ITERATOR_SIZE));
+        let mut a = BooleanChunked::from_slice("test", &generate_boolean_vec(SKIP_ITERATOR_SIZE));
+        let a_b = BooleanChunked::from_slice("test", &generate_boolean_vec(SKIP_ITERATOR_SIZE));
         a.append(&a_b);
         a
     });
 
     impl_test_iter_skip!(bool_iter_many_chunk_null_check_skip, 18, None, None, {
-        let mut a = BooleanChunked::new_from_opt_slice(
-            "test",
-            &generate_opt_boolean_vec(SKIP_ITERATOR_SIZE),
-        );
-        let a_b = BooleanChunked::new_from_opt_slice(
-            "test",
-            &generate_opt_boolean_vec(SKIP_ITERATOR_SIZE),
-        );
+        let mut a = BooleanChunked::new("test", &generate_opt_boolean_vec(SKIP_ITERATOR_SIZE));
+        let a_b = BooleanChunked::new("test", &generate_opt_boolean_vec(SKIP_ITERATOR_SIZE));
         a.append(&a_b);
         a
     });
