@@ -1,18 +1,49 @@
 #[cfg(feature = "object")]
 use crate::chunked_array::object::builder::ObjectChunkedBuilder;
 use crate::prelude::*;
+use crate::utils::slice_offsets;
 #[cfg(feature = "object")]
 use arrow::array::Array;
 use arrow::compute::concatenate;
-use itertools::Itertools;
-#[cfg(feature = "dtype-categorical")]
-use std::ops::Deref;
 
-pub trait ChunkOps {
-    /// Aggregate to contiguous memory.
-    fn rechunk(&self) -> Self
-    where
-        Self: std::marker::Sized;
+#[inline]
+fn slice(
+    chunks: &[ArrayRef],
+    offset: i64,
+    slice_length: usize,
+    own_length: usize,
+) -> Vec<ArrayRef> {
+    let mut new_chunks = Vec::with_capacity(1);
+    let (raw_offset, slice_len) = slice_offsets(offset, slice_length, own_length);
+
+    let mut remaining_length = slice_len;
+    let mut remaining_offset = raw_offset;
+
+    for chunk in chunks {
+        let chunk_len = chunk.len();
+        if remaining_offset > 0 && remaining_offset >= chunk_len {
+            remaining_offset -= chunk_len;
+            continue;
+        }
+        let take_len = if remaining_length + remaining_offset > chunk_len {
+            chunk_len - remaining_offset
+        } else {
+            remaining_length
+        };
+
+        debug_assert!(remaining_offset + take_len <= chunk.len());
+        unsafe {
+            // Safety:
+            // this function ensures the slices are in bounds
+            new_chunks.push(chunk.slice_unchecked(remaining_offset, take_len).into());
+        }
+        remaining_length -= take_len;
+        remaining_offset = 0;
+        if remaining_length == 0 {
+            break;
+        }
+    }
+    new_chunks
 }
 
 impl<T> ChunkOps for ChunkedArray<T>
@@ -24,13 +55,21 @@ where
             self.clone()
         } else {
             let chunks = vec![concatenate::concatenate(
-                self.chunks.iter().map(|a| &**a).collect_vec().as_slice(),
+                self.chunks
+                    .iter()
+                    .map(|a| &**a)
+                    .collect::<Vec<_>>()
+                    .as_slice(),
             )
             .unwrap()
             .into()];
-            ChunkedArray::new_from_chunks(self.name(), chunks)
+            ChunkedArray::from_chunks(self.name(), chunks)
         }
     }
+    #[inline]
+    fn slice(&self, offset: i64, length: usize) -> Self {
+        self.copy_with_chunks(slice(&self.chunks, offset, length, self.len()))
+    }
 }
 
 impl ChunkOps for BooleanChunked {
@@ -39,13 +78,21 @@ impl ChunkOps for BooleanChunked {
             self.clone()
         } else {
             let chunks = vec![concatenate::concatenate(
-                self.chunks.iter().map(|a| &**a).collect_vec().as_slice(),
+                self.chunks
+                    .iter()
+                    .map(|a| &**a)
+                    .collect::<Vec<_>>()
+                    .as_slice(),
             )
             .unwrap()
             .into()];
-            ChunkedArray::new_from_chunks(self.name(), chunks)
+            ChunkedArray::from_chunks(self.name(), chunks)
         }
     }
+    #[inline]
+    fn slice(&self, offset: i64, length: usize) -> Self {
+        self.copy_with_chunks(slice(&self.chunks, offset, length, self.len()))
+    }
 }
 
 impl ChunkOps for Utf8Chunked {
@@ -54,25 +101,20 @@ impl ChunkOps for Utf8Chunked {
             self.clone()
         } else {
             let chunks = vec![concatenate::concatenate(
-                self.chunks.iter().map(|a| &**a).collect_vec().as_slice(),
+                self.chunks
+                    .iter()
+                    .map(|a| &**a)
+                    .collect::<Vec<_>>()
+                    .as_slice(),
             )
             .unwrap()
             .into()];
-            ChunkedArray::new_from_chunks(self.name(), chunks)
+            ChunkedArray::from_chunks(self.name(), chunks)
         }
     }
-}
-
-#[cfg(feature = "dtype-categorical")]
-impl ChunkOps for CategoricalChunked {
-    fn rechunk(&self) -> Self
-    where
-        Self: std::marker::Sized,
-    {
-        let mut out: CategoricalChunked = self.deref().rechunk().into();
-        let cat_map = self.categorical_map.clone();
-        out.categorical_map = cat_map;
-        out
+    #[inline]
+    fn slice(&self, offset: i64, length: usize) -> Self {
+        self.copy_with_chunks(slice(&self.chunks, offset, length, self.len()))
     }
 }
 
@@ -82,17 +124,25 @@ impl ChunkOps for ListChunked {
             self.clone()
         } else {
             let chunks = vec![concatenate::concatenate(
-                self.chunks.iter().map(|a| &**a).collect_vec().as_slice(),
+                self.chunks
+                    .iter()
+                    .map(|a| &**a)
+                    .collect::<Vec<_>>()
+                    .as_slice(),
             )
             .unwrap()
             .into()];
-            let mut ca = ListChunked::new_from_chunks(self.name(), chunks);
+            let mut ca = ListChunked::from_chunks(self.name(), chunks);
             if self.can_fast_explode() {
                 ca.set_fast_explode()
             }
             ca
         }
     }
+    #[inline]
+    fn slice(&self, offset: i64, length: usize) -> Self {
+        self.copy_with_chunks(slice(&self.chunks, offset, length, self.len()))
+    }
 }
 
 #[cfg(feature = "object")]
@@ -132,6 +182,10 @@ where
             builder.finish()
         }
     }
+    #[inline]
+    fn slice(&self, offset: i64, length: usize) -> Self {
+        self.copy_with_chunks(slice(&self.chunks, offset, length, self.len()))
+    }
 }
 
 #[cfg(test)]
@@ -142,10 +196,10 @@ mod test {
     #[cfg(feature = "dtype-categorical")]
     fn test_categorical_map_after_rechunk() {
         let s = Series::new("", &["foo", "bar", "spam"]);
-        let mut a = s.cast(&DataType::Categorical).unwrap();
+        let mut a = s.cast(&DataType::Categorical(None)).unwrap();
 
         a.append(&a.slice(0, 2)).unwrap();
-        a.rechunk();
-        assert!(a.categorical().unwrap().categorical_map.is_some());
+        let a = a.rechunk();
+        assert!(a.categorical().unwrap().get_rev_map().len() > 0);
     }
 }
