@@ -8,14 +8,13 @@ pub use self::take::*;
 use crate::chunked_array::object::ObjectType;
 use crate::prelude::*;
 use arrow::buffer::Buffer;
-#[cfg(feature = "dtype-categorical")]
-use std::ops::Deref;
+use polars_arrow::prelude::QuantileInterpolOptions;
 
 #[cfg(feature = "abs")]
 mod abs;
 pub(crate) mod aggregate;
-mod any_value;
-mod append;
+pub(crate) mod any_value;
+pub(crate) mod append;
 mod apply;
 mod bit_repr;
 pub(crate) mod chunkops;
@@ -26,6 +25,7 @@ mod concat_str;
 mod cum_agg;
 pub(crate) mod downcast;
 pub(crate) mod explode;
+mod extend;
 mod fill_null;
 mod filter;
 pub mod full;
@@ -33,6 +33,7 @@ pub mod full;
 mod interpolate;
 #[cfg(feature = "is_in")]
 mod is_in;
+mod len;
 mod peaks;
 #[cfg(feature = "repeat_by")]
 mod repeat_by;
@@ -57,6 +58,7 @@ pub trait ToList<T: PolarsDataType> {
 
 #[cfg(feature = "interpolate")]
 pub trait Interpolate {
+    #[must_use]
     fn interpolate(&self) -> Self;
 }
 
@@ -136,7 +138,11 @@ pub trait ChunkBytes {
 /// This likely is a bit slower than ChunkWindow
 #[cfg(feature = "rolling_window")]
 pub trait ChunkRollApply {
-    fn rolling_apply(&self, _window_size: usize, _f: &dyn Fn(&Series) -> Series) -> Result<Self>
+    fn rolling_apply(
+        &self,
+        _f: &dyn Fn(&Series) -> Series,
+        _options: RollingOptions,
+    ) -> Result<Series>
     where
         Self: Sized,
     {
@@ -198,6 +204,7 @@ pub trait ChunkTake {
     /// # Safety
     ///
     /// Doesn't do any bound checking.
+    #[must_use]
     unsafe fn take_unchecked<I, INulls>(&self, indices: TakeIdx<I, INulls>) -> Self
     where
         Self: std::marker::Sized,
@@ -224,7 +231,7 @@ pub trait ChunkSet<'a, A, B> {
     ///
     /// ```rust
     /// # use polars_core::prelude::*;
-    /// let ca = UInt32Chunked::new_from_slice("a", &[1, 2, 3]);
+    /// let ca = UInt32Chunked::new("a", &[1, 2, 3]);
     /// let new = ca.set_at_idx(vec![0, 1], Some(10)).unwrap();
     ///
     /// assert_eq!(Vec::from(&new), &[Some(10), Some(10), Some(3)]);
@@ -243,7 +250,7 @@ pub trait ChunkSet<'a, A, B> {
     ///
     /// ```rust
     /// # use polars_core::prelude::*;
-    /// let ca = Int32Chunked::new_from_slice("a", &[1, 2, 3]);
+    /// let ca = Int32Chunked::new("a", &[1, 2, 3]);
     /// let new = ca.set_at_idx_with(vec![0, 1], |opt_v| opt_v.map(|v| v - 5)).unwrap();
     ///
     /// assert_eq!(Vec::from(&new), &[Some(-4), Some(-3), Some(3)]);
@@ -258,8 +265,8 @@ pub trait ChunkSet<'a, A, B> {
     ///
     /// ```rust
     /// # use polars_core::prelude::*;
-    /// let ca = Int32Chunked::new_from_slice("a", &[1, 2, 3]);
-    /// let mask = BooleanChunked::new_from_slice("mask", &[false, true, false]);
+    /// let ca = Int32Chunked::new("a", &[1, 2, 3]);
+    /// let mask = BooleanChunked::new("mask", &[false, true, false]);
     /// let new = ca.set(&mask, Some(5)).unwrap();
     /// assert_eq!(Vec::from(&new), &[Some(1), Some(5), Some(3)]);
     /// ```
@@ -273,8 +280,8 @@ pub trait ChunkSet<'a, A, B> {
     ///
     /// ```rust
     /// # use polars_core::prelude::*;
-    /// let ca = UInt32Chunked::new_from_slice("a", &[1, 2, 3]);
-    /// let mask = BooleanChunked::new_from_slice("mask", &[false, true, false]);
+    /// let ca = UInt32Chunked::new("a", &[1, 2, 3]);
+    /// let mask = BooleanChunked::new("mask", &[false, true, false]);
     /// let new = ca.set_with(&mask, |opt_v| opt_v.map(
     ///     |v| v * 2
     /// )).unwrap();
@@ -323,6 +330,7 @@ pub trait ChunkApply<'a, A, B> {
     ///     ca.apply(|v| v * 2)
     /// }
     /// ```
+    #[must_use]
     fn apply<F>(&'a self, f: F) -> Self
     where
         F: Fn(A) -> B + Copy;
@@ -333,16 +341,19 @@ pub trait ChunkApply<'a, A, B> {
         Self: Sized;
 
     /// Apply a closure elementwise including null values.
+    #[must_use]
     fn apply_on_opt<F>(&'a self, f: F) -> Self
     where
         F: Fn(Option<A>) -> Option<B> + Copy;
 
     /// Apply a closure elementwise. The closure gets the index of the element as first argument.
+    #[must_use]
     fn apply_with_idx<F>(&'a self, f: F) -> Self
     where
         F: Fn((usize, A)) -> B + Copy;
 
     /// Apply a closure elementwise. The closure gets the index of the element as first argument.
+    #[must_use]
     fn apply_with_idx_on_opt<F>(&'a self, f: F) -> Self
     where
         F: Fn((usize, Option<A>)) -> Option<B> + Copy;
@@ -376,16 +387,18 @@ pub trait ChunkAgg<T> {
     fn mean(&self) -> Option<f64> {
         None
     }
+}
 
+/// Quantile and median aggregation
+pub trait ChunkQuantile<T> {
     /// Returns the mean value in the array.
     /// Returns `None` if the array is empty or only contains null values.
-    fn median(&self) -> Option<f64> {
+    fn median(&self) -> Option<T> {
         None
     }
-
     /// Aggregate a given quantile of the ChunkedArray.
     /// Returns `None` if the array is empty or only contains null values.
-    fn quantile(&self, _quantile: f64) -> Result<Option<T>> {
+    fn quantile(&self, _quantile: f64, _interpol: QuantileInterpolOptions) -> Result<Option<T>> {
         Ok(None)
     }
 }
@@ -450,7 +463,7 @@ pub trait ChunkUnique<T> {
 
     /// Get first index of the unique values in a `ChunkedArray`.
     /// This Vec is sorted.
-    fn arg_unique(&self) -> Result<UInt32Chunked>;
+    fn arg_unique(&self) -> Result<IdxCa>;
 
     /// Number of unique values in the `ChunkedArray`
     fn n_unique(&self) -> Result<usize> {
@@ -471,13 +484,6 @@ pub trait ChunkUnique<T> {
         ))
     }
 
-    /// Count the unique values.
-    fn value_counts(&self) -> Result<DataFrame> {
-        Err(PolarsError::InvalidOperation(
-            "is_duplicated is not implemented for this dtype".into(),
-        ))
-    }
-
     /// The most occurring value(s). Can return multiple Values
     #[cfg(feature = "mode")]
     #[cfg_attr(docsrs, doc(cfg(feature = "mode")))]
@@ -496,30 +502,25 @@ pub trait ToDummies<T>: ChunkUnique<T> {
     }
 }
 
-#[derive(Default)]
+#[derive(Default, Copy, Clone, Eq, PartialEq, Debug)]
 pub struct SortOptions {
-    descending: bool,
-    nulls_last: bool,
+    pub descending: bool,
+    pub nulls_last: bool,
 }
 
 /// Sort operations on `ChunkedArray`.
 pub trait ChunkSort<T> {
     #[allow(unused_variables)]
-    fn sort_with(&self, options: SortOptions) -> ChunkedArray<T> {
-        unimplemented!()
-    }
+    fn sort_with(&self, options: SortOptions) -> ChunkedArray<T>;
 
     /// Returned a sorted `ChunkedArray`.
     fn sort(&self, reverse: bool) -> ChunkedArray<T>;
 
-    /// Sort this array in place.
-    fn sort_in_place(&mut self, reverse: bool);
-
     /// Retrieve the indexes needed to sort this array.
-    fn argsort(&self, reverse: bool) -> UInt32Chunked;
+    fn argsort(&self, options: SortOptions) -> IdxCa;
 
     /// Retrieve the indexes need to sort this and the other arrays.
-    fn argsort_multiple(&self, _other: &[Series], _reverse: &[bool]) -> Result<UInt32Chunked> {
+    fn argsort_multiple(&self, _other: &[Series], _reverse: &[bool]) -> Result<IdxCa> {
         Err(PolarsError::InvalidOperation(
             "argsort_multiple not implemented for this dtype".into(),
         ))
@@ -594,8 +595,8 @@ pub trait ChunkFilter<T> {
     ///
     /// ```rust
     /// # use polars_core::prelude::*;
-    /// let array = Int32Chunked::new_from_slice("array", &[1, 2, 3]);
-    /// let mask = BooleanChunked::new_from_slice("mask", &[true, false, true]);
+    /// let array = Int32Chunked::new("array", &[1, 2, 3]);
+    /// let mask = BooleanChunked::new("mask", &[true, false, true]);
     ///
     /// let filtered = array.filter(&mask).unwrap();
     /// assert_eq!(Vec::from(&filtered), [Some(1), Some(3)])
@@ -643,14 +644,6 @@ impl ChunkExpandAtIndex<Utf8Type> for Utf8Chunked {
     }
 }
 
-#[cfg(feature = "dtype-categorical")]
-impl ChunkExpandAtIndex<CategoricalType> for CategoricalChunked {
-    fn expand_at_index(&self, index: usize, length: usize) -> CategoricalChunked {
-        let ca: CategoricalChunked = self.deref().expand_at_index(index, length).into();
-        ca.set_state(self)
-    }
-}
-
 impl ChunkExpandAtIndex<ListType> for ListChunked {
     fn expand_at_index(&self, index: usize, length: usize) -> ListChunked {
         let opt_val = self.get(index);
@@ -693,14 +686,12 @@ pub trait ChunkZip<T> {
 /// Apply kernels on the arrow array chunks in a ChunkedArray.
 pub trait ChunkApplyKernel<A: Array> {
     /// Apply kernel and return result as a new ChunkedArray.
-    fn apply_kernel<F>(&self, f: F) -> Self
-    where
-        F: Fn(&A) -> ArrayRef;
+    #[must_use]
+    fn apply_kernel(&self, f: &dyn Fn(&A) -> ArrayRef) -> Self;
 
     /// Apply a kernel that outputs an array of different type.
-    fn apply_kernel_cast<F, S>(&self, f: F) -> ChunkedArray<S>
+    fn apply_kernel_cast<S>(&self, f: &dyn Fn(&A) -> ArrayRef) -> ChunkedArray<S>
     where
-        F: Fn(&A) -> ArrayRef,
         S: PolarsDataType;
 }
 
@@ -744,7 +735,7 @@ pub trait ArgAgg {
 #[cfg_attr(docsrs, doc(cfg(feature = "repeat_by")))]
 pub trait RepeatBy {
     /// Repeat the values `n` times, where `n` is determined by the values in `by`.
-    fn repeat_by(&self, _by: &UInt32Chunked) -> ListChunked {
+    fn repeat_by(&self, _by: &IdxCa) -> ListChunked {
         unimplemented!()
     }
 }
@@ -781,3 +772,65 @@ pub trait StrConcat {
     /// * `delimiter` - A string that will act as delimiter between values.
     fn str_concat(&self, delimiter: &str) -> Utf8Chunked;
 }
+
+pub trait ChunkLen {
+    /// Get the length of the ChunkedArray
+    fn len(&self) -> usize;
+
+    /// Check if ChunkedArray is empty.
+    fn is_empty(&self) -> bool {
+        self.len() == 0
+    }
+}
+
+pub trait ChunkOps: ChunkLen {
+    /// Aggregate to contiguous memory.
+    #[must_use]
+    fn rechunk(&self) -> Self
+    where
+        Self: std::marker::Sized;
+
+    /// Slice the array. The chunks are reallocated the underlying data slices are zero copy.
+    ///
+    /// When offset is negative it will be counted from the end of the array.
+    /// This method will never error,
+    /// and will slice the best match when offset, or length is out of bounds
+    #[must_use]
+    fn slice(&self, offset: i64, length: usize) -> Self
+    where
+        Self: std::marker::Sized;
+
+    /// Take a view of top n elements
+    #[must_use]
+    fn limit(&self, num_elements: usize) -> Self
+    where
+        Self: Sized,
+    {
+        self.slice(0, num_elements)
+    }
+
+    /// Get the head of the ChunkedArray
+    #[must_use]
+    fn head(&self, length: Option<usize>) -> Self
+    where
+        Self: Sized,
+    {
+        match length {
+            Some(len) => self.slice(0, std::cmp::min(len, self.len())),
+            None => self.slice(0, std::cmp::min(10, self.len())),
+        }
+    }
+
+    /// Get the tail of the ChunkedArray
+    #[must_use]
+    fn tail(&self, length: Option<usize>) -> Self
+    where
+        Self: Sized,
+    {
+        let len = match length {
+            Some(len) => std::cmp::min(len, self.len()),
+            None => std::cmp::min(10, self.len()),
+        };
+        self.slice(-(len as i64), len)
+    }
+}
