@@ -3,9 +3,9 @@ use crate::prelude::{AnyValue, TimeUnit};
 #[cfg(feature = "dtype-time")]
 use arrow::temporal_conversions::time64ns_to_time;
 use arrow::temporal_conversions::{
-    timestamp_ms_to_datetime, timestamp_ns_to_datetime, MILLISECONDS,
+    timestamp_ms_to_datetime, timestamp_ns_to_datetime, timestamp_us_to_datetime, MILLISECONDS,
 };
-use polars_time::export::chrono::{NaiveDateTime, NaiveTime};
+use chrono::{NaiveDateTime, NaiveTime};
 
 /// Number of seconds in a day
 pub(crate) const SECONDS_IN_DAY: i64 = 86_400;
@@ -18,6 +18,7 @@ impl From<&AnyValue<'_>> for NaiveDateTime {
             #[cfg(feature = "dtype-datetime")]
             AnyValue::Datetime(v, tu, _) => match tu {
                 TimeUnit::Nanoseconds => timestamp_ns_to_datetime(*v),
+                TimeUnit::Microseconds => timestamp_us_to_datetime(*v),
                 TimeUnit::Milliseconds => timestamp_ms_to_datetime(*v),
             },
             _ => panic!("can only convert date/datetime to NaiveDateTime"),
@@ -37,22 +38,32 @@ impl From<&AnyValue<'_>> for NaiveTime {
 
 // Used by lazy for literal conversion
 #[cfg(feature = "private")]
-pub fn naive_datetime_to_datetime_ns(v: &NaiveDateTime) -> i64 {
+pub fn datetime_to_timestamp_ns(v: NaiveDateTime) -> i64 {
     v.timestamp_nanos()
 }
 
 // Used by lazy for literal conversion
 #[cfg(feature = "private")]
-pub fn naive_datetime_to_datetime_ms(v: &NaiveDateTime) -> i64 {
+pub fn datetime_to_timestamp_ms(v: NaiveDateTime) -> i64 {
     v.timestamp_millis()
 }
 
-pub(crate) fn naive_datetime_to_date(v: &NaiveDateTime) -> i32 {
-    (naive_datetime_to_datetime_ms(v) / (MILLISECONDS * SECONDS_IN_DAY)) as i32
+// Used by lazy for literal conversion
+#[cfg(feature = "private")]
+pub fn datetime_to_timestamp_us(v: NaiveDateTime) -> i64 {
+    let us = v.timestamp() * 1_000_000;
+    us + v.timestamp_subsec_micros() as i64
+}
+
+pub(crate) fn naive_datetime_to_date(v: NaiveDateTime) -> i32 {
+    (datetime_to_timestamp_ms(v) / (MILLISECONDS * SECONDS_IN_DAY)) as i32
 }
 
 pub(crate) fn naive_date_to_date(nd: NaiveDate) -> i32 {
     let nt = NaiveTime::from_hms(0, 0, 0);
     let ndt = NaiveDateTime::new(nd, nt);
-    naive_datetime_to_date(&ndt)
+    naive_datetime_to_date(ndt)
 }
+pub(crate) const NS_IN_DAY: i64 = 86_400_000_000_000;
+pub(crate) const US_IN_DAY: i64 = 86_400_000_000;
+pub(crate) const MS_IN_DAY: i64 = 86_400_000;
