@@ -1,10 +1,9 @@
 use self::hashing::*;
-use crate::chunked_array::builder::PrimitiveChunkedBuilder;
 use crate::prelude::*;
 #[cfg(feature = "groupby_list")]
 use crate::utils::Wrap;
 use crate::utils::{
-    accumulate_dataframes_vertical, copy_from_slice_unchecked, set_partition_size, split_ca, NoNull,
+    accumulate_dataframes_vertical, copy_from_slice_unchecked, set_partition_size, split_offsets,
 };
 use crate::vector_hasher::{get_null_hash_value, AsU64, StrHash};
 use crate::POOL;
@@ -15,307 +14,28 @@ use polars_arrow::prelude::QuantileInterpolOptions;
 use rayon::prelude::*;
 use std::fmt::Debug;
 use std::hash::Hash;
-#[cfg(feature = "dtype-categorical")]
-use std::ops::Deref;
 
 pub mod aggregations;
-#[cfg(feature = "dynamic_groupby")]
-mod dynamic;
 pub(crate) mod hashing;
+mod into_groups;
 #[cfg(feature = "rows")]
 pub(crate) mod pivot;
+mod proxy;
+
 #[cfg(feature = "rows")]
 pub use pivot::PivotAgg;
-#[cfg(not(feature = "dynamic_groupby"))]
-#[derive(Clone, Debug)]
-pub struct DynamicGroupOptions {
-    pub index_column: String,
-}
 
-pub type GroupTuples = Vec<(u32, Vec<u32>)>;
-pub type GroupedMap<T> = HashMap<T, Vec<u32>, RandomState>;
-
-#[cfg(feature = "dynamic_groupby")]
-pub use dynamic::*;
-
-/// Used to create the tuples for a groupby operation.
-pub trait IntoGroupTuples {
-    /// Create the tuples need for a groupby operation.
-    ///     * The first value in the tuple is the first index of the group.
-    ///     * The second value in the tuple is are the indexes of the groups including the first value.
-    fn group_tuples(&self, _multithreaded: bool) -> GroupTuples {
-        unimplemented!()
-    }
-}
-
-fn group_multithreaded<T>(ca: &ChunkedArray<T>) -> bool {
-    // TODO! change to something sensible
-    ca.len() > 1000
-}
-
-fn num_group_tuples<T>(ca: &ChunkedArray<T>, multithreaded: bool) -> GroupTuples
-where
-    T: PolarsIntegerType,
-    T::Native: Hash + Eq + Send + AsU64,
-    Option<T::Native>: AsU64,
-{
-    #[cfg(feature = "dtype-categorical")]
-    let group_size_hint = if let Some(m) = &ca.categorical_map {
-        ca.len() / m.len()
-    } else {
-        0
-    };
-    #[cfg(not(feature = "dtype-categorical"))]
-    let group_size_hint = 0;
-    if multithreaded && group_multithreaded(ca) {
-        let n_partitions = set_partition_size() as u64;
-
-        // use the arrays as iterators
-        if ca.chunks.len() == 1 {
-            if !ca.has_validity() {
-                let keys = vec![ca.cont_slice().unwrap()];
-                groupby_threaded_num(keys, group_size_hint, n_partitions)
-            } else {
-                let keys = ca
-                    .downcast_iter()
-                    .map(|arr| arr.into_iter().map(|x| x.copied()).collect::<Vec<_>>())
-                    .collect::<Vec<_>>();
-                groupby_threaded_num(keys, group_size_hint, n_partitions)
-            }
-            // use the polars-iterators
-        } else if !ca.has_validity() {
-            let keys = vec![ca.into_no_null_iter().collect::<Vec<_>>()];
-            groupby_threaded_num(keys, group_size_hint, n_partitions)
-        } else {
-            let keys = vec![ca.into_iter().collect::<Vec<_>>()];
-            groupby_threaded_num(keys, group_size_hint, n_partitions)
-        }
-    } else if !ca.has_validity() {
-        groupby(ca.into_no_null_iter())
-    } else {
-        groupby(ca.into_iter())
-    }
-}
-
-impl<T> IntoGroupTuples for ChunkedArray<T>
-where
-    T: PolarsNumericType,
-    T::Native: NumCast,
-{
-    fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        match self.dtype() {
-            DataType::UInt64 => {
-                // convince the compiler that we are this type.
-                let ca: &UInt64Chunked = unsafe {
-                    &*(self as *const ChunkedArray<T> as *const ChunkedArray<UInt64Type>)
-                };
-                num_group_tuples(ca, multithreaded)
-            }
-            DataType::UInt32 => {
-                // convince the compiler that we are this type.
-                let ca: &UInt32Chunked = unsafe {
-                    &*(self as *const ChunkedArray<T> as *const ChunkedArray<UInt32Type>)
-                };
-                num_group_tuples(ca, multithreaded)
-            }
-            DataType::Int64 | DataType::Float64 => {
-                let ca = self.bit_repr_large();
-                num_group_tuples(&ca, multithreaded)
-            }
-            DataType::Int32 | DataType::Float32 => {
-                let ca = self.bit_repr_small();
-                num_group_tuples(&ca, multithreaded)
-            }
-            _ => {
-                let ca = self.cast(&DataType::UInt32).unwrap();
-                let ca = ca.u32().unwrap();
-                num_group_tuples(ca, multithreaded)
-            }
-        }
-    }
-}
-impl IntoGroupTuples for BooleanChunked {
-    fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        let ca = self.cast(&DataType::UInt32).unwrap();
-        let ca = ca.u32().unwrap();
-        ca.group_tuples(multithreaded)
-    }
-}
-
-impl IntoGroupTuples for Utf8Chunked {
-    fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        let hb = RandomState::default();
-        let null_h = get_null_hash_value(hb.clone());
-
-        if multithreaded {
-            let n_partitions = set_partition_size();
-
-            let split = split_ca(self, n_partitions).unwrap();
-
-            let str_hashes = POOL.install(|| {
-                split
-                    .par_iter()
-                    .map(|ca| {
-                        ca.into_iter()
-                            .map(|opt_s| {
-                                let hash = match opt_s {
-                                    Some(s) => str::get_hash(s, &hb),
-                                    None => null_h,
-                                };
-                                StrHash::new(opt_s, hash)
-                            })
-                            .collect::<Vec<_>>()
-                    })
-                    .collect::<Vec<_>>()
-            });
-            groupby_threaded_num(str_hashes, 0, n_partitions as u64)
-        } else {
-            let str_hashes = self
-                .into_iter()
-                .map(|opt_s| {
-                    let hash = match opt_s {
-                        Some(s) => str::get_hash(s, &hb),
-                        None => null_h,
-                    };
-                    StrHash::new(opt_s, hash)
-                })
-                .collect::<Vec<_>>();
-            groupby(str_hashes.iter())
-        }
-    }
-}
-
-#[cfg(feature = "dtype-categorical")]
-impl IntoGroupTuples for CategoricalChunked {
-    fn group_tuples(&self, multithreaded: bool) -> GroupTuples {
-        self.deref().group_tuples(multithreaded)
-    }
-}
-
-impl IntoGroupTuples for ListChunked {
-    #[cfg(feature = "groupby_list")]
-    fn group_tuples(&self, _multithreaded: bool) -> GroupTuples {
-        groupby(self.into_iter().map(|opt_s| opt_s.map(Wrap)))
-    }
-}
-
-#[cfg(feature = "object")]
-impl<T> IntoGroupTuples for ObjectChunked<T>
-where
-    T: PolarsObject,
-{
-    fn group_tuples(&self, _multithreaded: bool) -> GroupTuples {
-        groupby(self.into_iter())
-    }
-}
-
-/// Used to tightly two 32 bit values and null information
-/// Only the bit values matter, not the meaning of the bits
-#[inline]
-fn pack_u32_tuples(opt_l: Option<u32>, opt_r: Option<u32>) -> [u8; 9] {
-    // 4 bytes for first value
-    // 4 bytes for second value
-    // last bytes' bits are used to indicate missing values
-    let mut val = [0u8; 9];
-    let s = &mut val;
-    match (opt_l, opt_r) {
-        (Some(l), Some(r)) => {
-            // write to first 4 places
-            unsafe { copy_from_slice_unchecked(&l.to_ne_bytes(), &mut s[..4]) }
-            // write to second chunk of 4 places
-            unsafe { copy_from_slice_unchecked(&r.to_ne_bytes(), &mut s[4..8]) }
-            // leave last byte as is
-        }
-        (Some(l), None) => {
-            unsafe { copy_from_slice_unchecked(&l.to_ne_bytes(), &mut s[..4]) }
-            // set right null bit
-            s[8] = 1;
-        }
-        (None, Some(r)) => {
-            unsafe { copy_from_slice_unchecked(&r.to_ne_bytes(), &mut s[4..8]) }
-            // set left null bit
-            s[8] = 1 << 1;
-        }
-        (None, None) => {
-            // set two null bits
-            s[8] = 3;
-        }
-    }
-    val
-}
-
-/// Used to tightly two 64 bit values and null information
-/// Only the bit values matter, not the meaning of the bits
-#[inline]
-fn pack_u64_tuples(opt_l: Option<u64>, opt_r: Option<u64>) -> [u8; 17] {
-    // 8 bytes for first value
-    // 8 bytes for second value
-    // last bytes' bits are used to indicate missing values
-    let mut val = [0u8; 17];
-    let s = &mut val;
-    match (opt_l, opt_r) {
-        (Some(l), Some(r)) => {
-            // write to first 4 places
-            unsafe { copy_from_slice_unchecked(&l.to_ne_bytes(), &mut s[..8]) }
-            // write to second chunk of 4 places
-            unsafe { copy_from_slice_unchecked(&r.to_ne_bytes(), &mut s[8..16]) }
-            // leave last byte as is
-        }
-        (Some(l), None) => {
-            unsafe { copy_from_slice_unchecked(&l.to_ne_bytes(), &mut s[..8]) }
-            // set right null bit
-            s[16] = 1;
-        }
-        (None, Some(r)) => {
-            unsafe { copy_from_slice_unchecked(&r.to_ne_bytes(), &mut s[8..16]) }
-            // set left null bit
-            s[16] = 1 << 1;
-        }
-        (None, None) => {
-            // set two null bits
-            s[16] = 3;
-        }
-    }
-    val
-}
-
-/// Used to tightly one 32 bit and a 64 bit valued type and null information
-/// Only the bit values matter, not the meaning of the bits
-#[inline]
-fn pack_u32_u64_tuples(opt_l: Option<u32>, opt_r: Option<u64>) -> [u8; 13] {
-    // 8 bytes for first value
-    // 8 bytes for second value
-    // last bytes' bits are used to indicate missing values
-    let mut val = [0u8; 13];
-    let s = &mut val;
-    match (opt_l, opt_r) {
-        (Some(l), Some(r)) => {
-            // write to first 4 places
-            unsafe { copy_from_slice_unchecked(&l.to_ne_bytes(), &mut s[..4]) }
-            // write to second chunk of 4 places
-            unsafe { copy_from_slice_unchecked(&r.to_ne_bytes(), &mut s[4..12]) }
-            // leave last byte as is
-        }
-        (Some(l), None) => {
-            unsafe { copy_from_slice_unchecked(&l.to_ne_bytes(), &mut s[..4]) }
-            // set right null bit
-            s[12] = 1;
-        }
-        (None, Some(r)) => {
-            unsafe { copy_from_slice_unchecked(&r.to_ne_bytes(), &mut s[4..12]) }
-            // set left null bit
-            s[12] = 1 << 1;
-        }
-        (None, None) => {
-            // set two null bits
-            s[12] = 3;
-        }
-    }
-    val
-}
+pub use into_groups::*;
+use polars_arrow::array::ValueSize;
+pub use proxy::*;
 
 impl DataFrame {
-    pub fn groupby_with_series(&self, by: Vec<Series>, multithreaded: bool) -> Result<GroupBy> {
+    pub fn groupby_with_series(
+        &self,
+        by: Vec<Series>,
+        multithreaded: bool,
+        sorted: bool,
+    ) -> Result<GroupBy> {
         macro_rules! finish_packed_bit_path {
             ($ca0:expr, $ca1:expr, $pack_fn:expr) => {{
                 let n_partitions = set_partition_size();
@@ -324,14 +44,14 @@ impl DataFrame {
                 // pack the bit values together and add a final byte that will be 0
                 // when there are no null values.
                 // otherwise we use two bits of this byte to represent null values.
-                let split_0 = split_ca(&$ca0, n_partitions).unwrap();
-                let split_1 = split_ca(&$ca1, n_partitions).unwrap();
+                let splits = split_offsets($ca0.len(), n_partitions);
 
                 let keys = POOL.install(|| {
-                    split_0
+                    splits
                         .into_par_iter()
-                        .zip(split_1.into_par_iter())
-                        .map(|(ca0, ca1)| {
+                        .map(|(offset, len)| {
+                            let ca0 = $ca0.slice(offset as i64, len);
+                            let ca1 = $ca1.slice(offset as i64, len);
                             ca0.into_iter()
                                 .zip(ca1.into_iter())
                                 .map(|(l, r)| $pack_fn(l, r))
@@ -343,7 +63,7 @@ impl DataFrame {
                 return Ok(GroupBy::new(
                     self,
                     by,
-                    groupby_threaded_num(keys, 0, n_partitions as u64),
+                    groupby_threaded_num(keys, 0, n_partitions as u64, sorted),
                     None,
                 ));
             }};
@@ -357,18 +77,18 @@ impl DataFrame {
 
         use DataType::*;
         // make sure that categorical and small integers are used as uint32 in value type
-        let keys_df = DataFrame::new(
+        let keys_df = DataFrame::new_no_checks(
             by.iter()
                 .map(|s| match s.dtype() {
-                    Categorical | Int8 | UInt8 | Int16 | UInt16 => {
-                        s.cast(&DataType::UInt32).unwrap()
-                    }
+                    Int8 | UInt8 | Int16 | UInt16 => s.cast(&DataType::UInt32).unwrap(),
+                    #[cfg(feature = "dtype-categorical")]
+                    Categorical(_) => s.cast(&DataType::UInt32).unwrap(),
                     Float32 => s.bit_repr_small().into_series(),
                     // otherwise we use the vec hash for float
                     Float64 => s.bit_repr_large().into_series(),
                     _ => {
                         // is date like
-                        if !s.is_numeric() && s.is_numeric_physical() {
+                        if !s.dtype().is_numeric() && s.is_numeric_physical() {
                             s.to_physical_repr().into_owned()
                         } else {
                             s.clone()
@@ -376,14 +96,16 @@ impl DataFrame {
                     }
                 })
                 .collect(),
-        )?;
+        );
+
+        let n_partitions = set_partition_size();
 
         let groups = match by.len() {
             1 => {
                 let series = &by[0];
-                series.group_tuples(multithreaded)
+                series.group_tuples(multithreaded, sorted)
             }
-            _ => {
+            2 => {
                 // multiple keys is always multi-threaded
                 // reduce code paths
                 let s0 = &keys_df.get_columns()[0];
@@ -391,7 +113,7 @@ impl DataFrame {
 
                 // fast path for numeric data
                 // uses the bit values to tightly pack those into arrays.
-                if by.len() == 2 && s0.is_numeric() && s1.is_numeric() {
+                if s0.dtype().is_numeric() && s1.dtype().is_numeric() {
                     match (s0.bit_repr_is_large(), s1.bit_repr_is_large()) {
                         (false, false) => {
                             let ca0 = s0.bit_repr_small();
@@ -416,11 +138,22 @@ impl DataFrame {
                             finish_packed_bit_path!(ca0, ca1, pack_u32_u64_tuples)
                         }
                     }
-                }
+                } else if matches!((s0.dtype(), s1.dtype()), (DataType::Utf8, DataType::Utf8)) {
+                    let lhs = s0.utf8().unwrap();
+                    let rhs = s1.utf8().unwrap();
 
-                let n_partitions = set_partition_size();
-                groupby_threaded_multiple_keys_flat(keys_df, n_partitions)
+                    // arbitrarily chosen bound, if avg no of bytes to encode is larger than this
+                    // value we fall back to default groupby
+                    if (lhs.get_values_size() + rhs.get_values_size()) / lhs.len() < 128 {
+                        pack_utf8_columns(lhs, rhs, n_partitions, sorted)
+                    } else {
+                        groupby_threaded_multiple_keys_flat(keys_df, n_partitions, sorted)
+                    }
+                } else {
+                    groupby_threaded_multiple_keys_flat(keys_df, n_partitions, sorted)
+                }
             }
+            _ => groupby_threaded_multiple_keys_flat(keys_df, n_partitions, sorted),
         };
         Ok(GroupBy::new(self, by, groups, None))
     }
@@ -443,7 +176,7 @@ impl DataFrame {
         S: AsRef<str>,
     {
         let selected_keys = self.select_series(by)?;
-        self.groupby_with_series(selected_keys, true)
+        self.groupby_with_series(selected_keys, true, false)
     }
 
     /// Group DataFrame using a Series column.
@@ -453,9 +186,8 @@ impl DataFrame {
         I: IntoIterator<Item = S>,
         S: AsRef<str>,
     {
-        let mut gb = self.groupby(by)?;
-        gb.groups.sort_unstable_by_key(|t| t.0);
-        Ok(gb)
+        let selected_keys = self.select_series(by)?;
+        self.groupby_with_series(selected_keys, true, true)
     }
 }
 
@@ -513,7 +245,7 @@ pub struct GroupBy<'df> {
     df: &'df DataFrame,
     pub(crate) selected_keys: Vec<Series>,
     // [first idx, [other idx]]
-    pub(crate) groups: GroupTuples,
+    pub(crate) groups: GroupsProxy,
     // columns selected for aggregation
     pub(crate) selected_agg: Option<Vec<String>>,
 }
@@ -522,7 +254,7 @@ impl<'df> GroupBy<'df> {
     pub fn new(
         df: &'df DataFrame,
         by: Vec<Series>,
-        groups: GroupTuples,
+        groups: GroupsProxy,
         selected_agg: Option<Vec<String>>,
     ) -> Self {
         GroupBy {
@@ -553,7 +285,7 @@ impl<'df> GroupBy<'df> {
     /// The Vec returned contains:
     ///     (first_idx, Vec<indexes>)
     ///     Where second value in the tuple is a vector with all matching indexes.
-    pub fn get_groups(&self) -> &GroupTuples {
+    pub fn get_groups(&self) -> &GroupsProxy {
         &self.groups
     }
 
@@ -561,10 +293,14 @@ impl<'df> GroupBy<'df> {
     /// The Vec returned contains:
     ///     (first_idx, Vec<indexes>)
     ///     Where second value in the tuple is a vector with all matching indexes.
-    pub fn get_groups_mut(&mut self) -> &mut GroupTuples {
+    pub fn get_groups_mut(&mut self) -> &mut GroupsProxy {
         &mut self.groups
     }
 
+    pub fn take_groups(self) -> GroupsProxy {
+        self.groups
+    }
+
     pub fn keys(&self) -> Vec<Series> {
         POOL.install(|| {
             self.selected_keys
@@ -573,7 +309,9 @@ impl<'df> GroupBy<'df> {
                     // Safety
                     // groupby indexes are in bound.
                     unsafe {
-                        s.take_iter_unchecked(&mut self.groups.iter().map(|(idx, _)| *idx as usize))
+                        s.take_iter_unchecked(
+                            &mut self.groups.idx_ref().iter().map(|(idx, _)| idx as usize),
+                        )
                     }
                 })
                 .collect()
@@ -972,15 +710,12 @@ impl<'df> GroupBy<'df> {
     /// ```
     pub fn count(&self) -> Result<DataFrame> {
         let (mut cols, agg_cols) = self.prepare_agg()?;
+
         for agg_col in agg_cols {
             let new_name = fmt_groupby_column(agg_col.name(), GroupByMethod::Count);
-            let mut builder =
-                PrimitiveChunkedBuilder::<UInt32Type>::new(&new_name, self.groups.len());
-            for (_first, idx) in &self.groups {
-                builder.append_value(idx.len() as u32);
-            }
-            let ca = builder.finish();
-            cols.push(ca.into_series())
+            let mut ca = self.groups.group_count();
+            ca.rename(&new_name);
+            cols.push(ca.into_series());
         }
         DataFrame::new(cols)
     }
@@ -1012,15 +747,7 @@ impl<'df> GroupBy<'df> {
     /// ```
     pub fn groups(&self) -> Result<DataFrame> {
         let mut cols = self.keys();
-
-        let mut column: ListChunked = self
-            .groups
-            .iter()
-            .map(|(_first, idx)| {
-                let ca: NoNull<UInt32Chunked> = idx.iter().map(|&v| v as u32).collect();
-                ca.into_inner().into_series()
-            })
-            .collect();
+        let mut column = self.groups.as_list_chunked();
         let new_name = fmt_groupby_column("", GroupByMethod::Groups);
         column.rename(&new_name);
         cols.push(column.into_series());
@@ -1115,14 +842,8 @@ impl<'df> GroupBy<'df> {
                         "var" => finish_agg_opt!(self, "{}_var", agg_var, agg_col, cols),
                         "count" => {
                             let new_name = format!("{}_count", agg_col.name());
-                            let mut builder = PrimitiveChunkedBuilder::<UInt32Type>::new(
-                                &new_name,
-                                self.groups.len(),
-                            );
-                            for (_first, idx) in &self.groups {
-                                builder.append_value(idx.len() as u32);
-                            }
-                            let ca = builder.finish();
+                            let mut ca = self.groups.group_count();
+                            ca.rename(&new_name);
                             cols.push(ca.into_series());
                         }
                         a => panic!("aggregation: {:?} is not supported", a),
@@ -1195,6 +916,7 @@ impl<'df> GroupBy<'df> {
         let df = self.prepare_apply()?;
         let dfs = self
             .get_groups()
+            .idx_ref()
             .par_iter()
             .map(|t| {
                 let sub_df = unsafe { df.take_iter_unchecked(t.1.iter().map(|i| *i as usize)) };
@@ -1215,6 +937,7 @@ impl<'df> GroupBy<'df> {
         let df = self.prepare_apply()?;
         let dfs = self
             .get_groups()
+            .idx_ref()
             .iter()
             .map(|t| {
                 let sub_df = unsafe { df.take_iter_unchecked(t.1.iter().map(|i| *i as usize)) };
@@ -1297,7 +1020,7 @@ mod test {
         let out = df.groupby_stable(["date"])?.select(["temp"]).count()?;
         assert_eq!(
             out.column("temp_count")?,
-            &Series::new("temp_count", [2u32, 2, 1])
+            &Series::new("temp_count", [2 as IdxSize, 2, 1])
         );
 
         // Select multiple
@@ -1442,7 +1165,7 @@ mod test {
         }
         .unwrap();
 
-        df.apply("foo", |s| s.cast(&DataType::Categorical).unwrap())
+        df.apply("foo", |s| s.cast(&DataType::Categorical(None)).unwrap())
             .unwrap();
 
         // check multiple keys and categorical
@@ -1483,14 +1206,10 @@ mod test {
             let ca = UInt32Chunked::new("", slice);
             let split = split_ca(&ca, 4).unwrap();
 
-            let mut a = groupby(ca.into_iter()).into_iter().collect::<Vec<_>>();
-            a.sort();
+            let a = groupby(ca.into_iter(), true).into_idx();
 
             let keys = split.iter().map(|ca| ca.cont_slice().unwrap()).collect();
-            let mut b = groupby_threaded_num(keys, 0, split.len() as u64)
-                .into_iter()
-                .collect::<Vec<_>>();
-            b.sort();
+            let b = groupby_threaded_num(keys, 0, split.len() as u64, true).into_idx();
 
             assert_eq!(a, b);
         }
@@ -1550,7 +1269,7 @@ mod test {
             "int" => [1, 2, 3, 1, 1]
         ]?;
 
-        df.try_apply("g", |s| s.cast(&DataType::Categorical))?;
+        df.try_apply("g", |s| s.cast(&DataType::Categorical(None)))?;
 
         let _ = df.groupby(["g"])?.sum()?;
         Ok(())
