@@ -1,4 +1,5 @@
 #![cfg_attr(docsrs, feature(doc_cfg))]
+extern crate core;
 #[macro_use]
 pub mod utils;
 pub mod chunked_array;
@@ -10,7 +11,9 @@ pub mod export;
 mod fmt;
 pub mod frame;
 pub mod functions;
+mod named_from;
 pub mod prelude;
+pub mod schema;
 #[cfg(feature = "serde")]
 #[cfg_attr(docsrs, doc(cfg(feature = "serde")))]
 pub mod serde;
