@@ -4,6 +4,7 @@ use super::Series;
 use crate::apply_method_numeric_series;
 use crate::prelude::*;
 use crate::series::arithmetic::coerce_lhs_rhs;
+use std::ops::Deref;
 
 macro_rules! impl_compare {
     ($self:expr, $rhs:expr, $method:ident) => {{
@@ -23,26 +24,34 @@ macro_rules! impl_compare {
             DataType::Int64 => lhs.i64().unwrap().$method(rhs.i64().unwrap()),
             DataType::Float32 => lhs.f32().unwrap().$method(rhs.f32().unwrap()),
             DataType::Float64 => lhs.f64().unwrap().$method(rhs.f64().unwrap()),
-            DataType::Date32 => lhs.date32().unwrap().$method(rhs.date32().unwrap()),
-            DataType::Date64 => lhs.date64().unwrap().$method(rhs.date64().unwrap()),
-            DataType::Time64(TimeUnit::Nanosecond) => lhs
-                .time64_nanosecond()
+            #[cfg(feature = "dtype-date")]
+            DataType::Date => lhs.date().unwrap().$method(rhs.date().unwrap().deref()),
+            #[cfg(feature = "dtype-time")]
+            DataType::Time => lhs.time().unwrap().$method(rhs.time().unwrap().deref()),
+            #[cfg(feature = "dtype-datetime")]
+            DataType::Datetime(_, _) => lhs
+                .datetime()
                 .unwrap()
-                .$method(rhs.time64_nanosecond().unwrap()),
-            DataType::Duration(TimeUnit::Nanosecond) => lhs
-                .duration_nanosecond()
+                .$method(rhs.datetime().unwrap().deref()),
+            #[cfg(feature = "dtype-duration")]
+            DataType::Duration(_) => lhs
+                .duration()
                 .unwrap()
-                .$method(rhs.duration_nanosecond().unwrap()),
-            DataType::Duration(TimeUnit::Millisecond) => lhs
-                .duration_millisecond()
-                .unwrap()
-                .$method(rhs.duration_millisecond().unwrap()),
+                .$method(rhs.duration().unwrap().deref()),
             DataType::List(_) => lhs.list().unwrap().$method(rhs.list().unwrap()),
+            #[cfg(feature = "dtype-categorical")]
+            DataType::Categorical(_) => lhs
+                .categorical()
+                .unwrap()
+                .logical()
+                .$method(rhs.categorical().unwrap().logical()),
+
             _ => unimplemented!(),
         }
     }};
 }
 
+#[cfg(feature = "dtype-categorical")]
 fn compare_cat_to_str_value<Compare>(
     cat: &Series,
     value: &str,
@@ -54,16 +63,17 @@ where
     Compare: Fn(&Series, u32) -> BooleanChunked,
 {
     let cat = cat.categorical().expect("should be categorical");
-    let cat_map = cat.get_categorical_map().unwrap();
+    let cat_map = cat.get_rev_map();
     match cat_map.find(value) {
         None => BooleanChunked::full(name, fill_value, cat.len()),
         Some(cat_idx) => {
-            let cat = cat.cast_with_dtype(&DataType::UInt32).unwrap();
+            let cat = cat.cast(&DataType::UInt32).unwrap();
             compare(&cat, cat_idx)
         }
     }
 }
 
+#[cfg(feature = "dtype-categorical")]
 fn compare_cat_to_str_series<Compare>(
     cat: &Series,
     string: &Series,
@@ -82,100 +92,144 @@ where
 
 impl ChunkCompare<&Series> for Series {
     fn eq_missing(&self, rhs: &Series) -> BooleanChunked {
-        impl_compare!(self, rhs, eq_missing)
+        #[cfg(feature = "dtype-categorical")]
+        use DataType::*;
+        match (self.dtype(), rhs.dtype(), self.len(), rhs.len()) {
+            #[cfg(feature = "dtype-categorical")]
+            (Categorical(_), Utf8, _, 1) => {
+                return compare_cat_to_str_series(
+                    self,
+                    rhs,
+                    self.name(),
+                    |s, idx| s.eq_missing(idx),
+                    false,
+                );
+            }
+            #[cfg(feature = "dtype-categorical")]
+            (Utf8, Categorical(_), 1, _) => {
+                return compare_cat_to_str_series(
+                    rhs,
+                    self,
+                    self.name(),
+                    |s, idx| s.eq_missing(idx),
+                    false,
+                );
+            }
+            _ => {
+                impl_compare!(self, rhs, eq_missing)
+            }
+        }
     }
 
     /// Create a boolean mask by checking for equality.
-    fn eq(&self, rhs: &Series) -> BooleanChunked {
+    fn equal(&self, rhs: &Series) -> BooleanChunked {
+        #[cfg(feature = "dtype-categorical")]
         use DataType::*;
-        match (self.dtype(), rhs.dtype(), self.len(), rhs.len()) {
-            (Categorical, Utf8, _, 1) => {
+        let mut out = match (self.dtype(), rhs.dtype(), self.len(), rhs.len()) {
+            #[cfg(feature = "dtype-categorical")]
+            (Categorical(_), Utf8, _, 1) => {
                 return compare_cat_to_str_series(
                     self,
                     rhs,
                     self.name(),
-                    |s, idx| s.eq(idx),
+                    |s, idx| s.equal(idx),
                     false,
                 );
             }
-            (Utf8, Categorical, 1, _) => {
+            #[cfg(feature = "dtype-categorical")]
+            (Utf8, Categorical(_), 1, _) => {
                 return compare_cat_to_str_series(
                     rhs,
                     self,
                     self.name(),
-                    |s, idx| s.eq(idx),
+                    |s, idx| s.equal(idx),
                     false,
                 );
             }
             _ => {
-                impl_compare!(self, rhs, eq)
+                impl_compare!(self, rhs, equal)
             }
-        }
+        };
+        out.rename(self.name());
+        out
     }
 
     /// Create a boolean mask by checking for inequality.
-    fn neq(&self, rhs: &Series) -> BooleanChunked {
+    fn not_equal(&self, rhs: &Series) -> BooleanChunked {
+        #[cfg(feature = "dtype-categorical")]
         use DataType::*;
-        match (self.dtype(), rhs.dtype(), self.len(), rhs.len()) {
-            (Categorical, Utf8, _, 1) => {
+        let mut out = match (self.dtype(), rhs.dtype(), self.len(), rhs.len()) {
+            #[cfg(feature = "dtype-categorical")]
+            (Categorical(_), Utf8, _, 1) => {
                 return compare_cat_to_str_series(
                     self,
                     rhs,
                     self.name(),
-                    |s, idx| s.neq(idx),
+                    |s, idx| s.not_equal(idx),
                     true,
                 );
             }
-            (Utf8, Categorical, 1, _) => {
+            #[cfg(feature = "dtype-categorical")]
+            (Utf8, Categorical(_), 1, _) => {
                 return compare_cat_to_str_series(
                     rhs,
                     self,
                     self.name(),
-                    |s, idx| s.neq(idx),
+                    |s, idx| s.not_equal(idx),
                     true,
                 );
             }
             _ => {
-                impl_compare!(self, rhs, neq)
+                impl_compare!(self, rhs, not_equal)
             }
-        }
+        };
+        out.rename(self.name());
+        out
     }
 
     /// Create a boolean mask by checking if self > rhs.
     fn gt(&self, rhs: &Series) -> BooleanChunked {
-        impl_compare!(self, rhs, gt)
+        let mut out = impl_compare!(self, rhs, gt);
+        out.rename(self.name());
+        out
     }
 
     /// Create a boolean mask by checking if self >= rhs.
     fn gt_eq(&self, rhs: &Series) -> BooleanChunked {
-        impl_compare!(self, rhs, gt_eq)
+        let mut out = impl_compare!(self, rhs, gt_eq);
+        out.rename(self.name());
+        out
     }
 
     /// Create a boolean mask by checking if self < rhs.
     fn lt(&self, rhs: &Series) -> BooleanChunked {
-        impl_compare!(self, rhs, lt)
+        let mut out = impl_compare!(self, rhs, lt);
+        out.rename(self.name());
+        out
     }
 
     /// Create a boolean mask by checking if self <= rhs.
     fn lt_eq(&self, rhs: &Series) -> BooleanChunked {
-        impl_compare!(self, rhs, lt_eq)
+        let mut out = impl_compare!(self, rhs, lt_eq);
+        out.rename(self.name());
+        out
     }
 }
 
 impl<Rhs> ChunkCompare<Rhs> for Series
 where
-    Rhs: NumComp,
+    Rhs: NumericNative,
 {
     fn eq_missing(&self, rhs: Rhs) -> BooleanChunked {
-        self.eq(rhs)
+        self.equal(rhs)
     }
 
-    fn eq(&self, rhs: Rhs) -> BooleanChunked {
-        apply_method_numeric_series!(self, eq, rhs)
+    fn equal(&self, rhs: Rhs) -> BooleanChunked {
+        apply_method_numeric_series!(self, equal, rhs)
     }
 
-    fn neq(&self, rhs: Rhs) -> BooleanChunked {
-        apply_method_numeric_series!(self, neq, rhs)
+    fn not_equal(&self, rhs: Rhs) -> BooleanChunked {
+        apply_method_numeric_series!(self, not_equal, rhs)
     }
 
     fn gt(&self, rhs: Rhs) -> BooleanChunked {
@@ -197,27 +251,33 @@ where
 
 impl ChunkCompare<&str> for Series {
     fn eq_missing(&self, rhs: &str) -> BooleanChunked {
-        self.eq(rhs)
+        self.equal(rhs)
     }
 
-    fn eq(&self, rhs: &str) -> BooleanChunked {
+    fn equal(&self, rhs: &str) -> BooleanChunked {
         use DataType::*;
         match self.dtype() {
-            Utf8 => self.utf8().unwrap().eq(rhs),
-            Categorical => {
-                compare_cat_to_str_value(self, rhs, self.name(), |lhs, idx| lhs.eq(idx), false)
+            Utf8 => self.utf8().unwrap().equal(rhs),
+            #[cfg(feature = "dtype-categorical")]
+            Categorical(_) => {
+                compare_cat_to_str_value(self, rhs, self.name(), |lhs, idx| lhs.equal(idx), false)
             }
             _ => BooleanChunked::full(self.name(), false, self.len()),
         }
     }
 
-    fn neq(&self, rhs: &str) -> BooleanChunked {
+    fn not_equal(&self, rhs: &str) -> BooleanChunked {
         use DataType::*;
         match self.dtype() {
-            Utf8 => self.utf8().unwrap().neq(rhs),
-            Categorical => {
-                compare_cat_to_str_value(self, rhs, self.name(), |lhs, idx| lhs.neq(idx), true)
-            }
+            Utf8 => self.utf8().unwrap().not_equal(rhs),
+            #[cfg(feature = "dtype-categorical")]
+            Categorical(_) => compare_cat_to_str_value(
+                self,
+                rhs,
+                self.name(),
+                |lhs, idx| lhs.not_equal(idx),
+                true,
+            ),
             _ => BooleanChunked::full(self.name(), false, self.len()),
         }
     }
