@@ -5,13 +5,19 @@ pub use crate::prelude::ChunkCompare;
 use crate::prelude::*;
 use arrow::array::ArrayRef;
 use polars_arrow::prelude::QuantileInterpolOptions;
-#[cfg(feature = "object")]
 use std::any::Any;
 use std::borrow::Cow;
 #[cfg(feature = "temporal")]
 use std::ops::Deref;
 use std::sync::Arc;
 
+#[derive(Debug, Copy, Clone)]
+pub enum IsSorted {
+    Ascending,
+    Descending,
+    Not,
+}
+
 macro_rules! invalid_operation {
     ($s:expr) => {
         Err(PolarsError::InvalidOperation(
@@ -35,7 +41,7 @@ macro_rules! invalid_operation_panic {
 pub(crate) mod private {
     use super::*;
     #[cfg(feature = "rows")]
-    use crate::frame::groupby::GroupTuples;
+    use crate::frame::groupby::GroupsProxy;
 
     use crate::chunked_array::ops::compare_inner::{PartialEqInner, PartialOrdInner};
     use ahash::RandomState;
@@ -138,11 +144,6 @@ pub(crate) mod private {
             panic!("operation cummin not supported for this dtype")
         }
 
-        #[cfg(feature = "asof_join")]
-        fn join_asof(&self, _other: &Series) -> Result<Vec<Option<u32>>> {
-            invalid_operation!(self)
-        }
-
         fn set_sorted(&mut self, _reverse: bool) {
             invalid_operation_panic!(self)
         }
@@ -169,66 +170,44 @@ pub(crate) mod private {
         fn vec_hash_combine(&self, _build_hasher: RandomState, _hashes: &mut [u64]) {
             invalid_operation_panic!(self)
         }
-        fn agg_mean(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_mean(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
-        fn agg_min(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_min(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
-        fn agg_max(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_max(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
         /// If the [`DataType`] is one of `{Int8, UInt8, Int16, UInt16}` the `Series` is
         /// first cast to `Int64` to prevent overflow issues.
-        fn agg_sum(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_sum(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
-        fn agg_std(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_std(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
-        fn agg_var(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_var(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
-        fn agg_first(&self, _groups: &[(u32, Vec<u32>)]) -> Series {
-            invalid_operation_panic!(self)
-        }
-        fn agg_last(&self, _groups: &[(u32, Vec<u32>)]) -> Series {
-            invalid_operation_panic!(self)
-        }
-        fn agg_n_unique(&self, _groups: &[(u32, Vec<u32>)]) -> Option<UInt32Chunked> {
-            None
-        }
-        fn agg_list(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_list(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
         fn agg_quantile(
             &self,
-            _groups: &[(u32, Vec<u32>)],
+            _groups: &GroupsProxy,
             _quantile: f64,
             _interpol: QuantileInterpolOptions,
         ) -> Option<Series> {
             None
         }
-        fn agg_median(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
+        fn agg_median(&self, _groups: &GroupsProxy) -> Option<Series> {
             None
         }
-        #[cfg(feature = "lazy")]
-        fn agg_valid_count(&self, _groups: &[(u32, Vec<u32>)]) -> Option<Series> {
-            None
-        }
-        fn hash_join_inner(&self, _other: &Series) -> Vec<(u32, u32)> {
-            invalid_operation_panic!(self)
-        }
-        fn hash_join_left(&self, _other: &Series) -> Vec<(u32, Option<u32>)> {
-            invalid_operation_panic!(self)
-        }
-        fn hash_join_outer(&self, _other: &Series) -> Vec<(Option<u32>, Option<u32>)> {
-            invalid_operation_panic!(self)
-        }
         fn zip_outer_join_column(
             &self,
             _right_column: &Series,
-            _opt_join_tuples: &[(Option<u32>, Option<u32>)],
+            _opt_join_tuples: &[(Option<IdxSize>, Option<IdxSize>)],
         ) -> Series {
             invalid_operation_panic!(self)
         }
@@ -248,14 +227,14 @@ pub(crate) mod private {
         fn remainder(&self, _rhs: &Series) -> Result<Series> {
             invalid_operation_panic!(self)
         }
-        fn group_tuples(&self, _multithreaded: bool) -> GroupTuples {
+        fn group_tuples(&self, _multithreaded: bool, _sorted: bool) -> GroupsProxy {
             invalid_operation_panic!(self)
         }
         fn zip_with_same_type(&self, _mask: &BooleanChunked, _other: &Series) -> Result<Series> {
             invalid_operation_panic!(self)
         }
         #[cfg(feature = "sort_multiple")]
-        fn argsort_multiple(&self, _by: &[Series], _reverse: &[bool]) -> Result<UInt32Chunked> {
+        fn argsort_multiple(&self, _by: &[Series], _reverse: &[bool]) -> Result<IdxCa> {
             Err(PolarsError::InvalidOperation(
                 "argsort_multiple is not implemented for this Series".into(),
             ))
@@ -266,6 +245,11 @@ pub(crate) mod private {
 pub trait SeriesTrait:
     Send + Sync + private::PrivateSeries + private::PrivateSeriesNumeric
 {
+    /// Check if [`Series`] is sorted.
+    fn is_sorted(&self) -> IsSorted {
+        IsSorted::Not
+    }
+
     #[cfg(feature = "interpolate")]
     #[cfg_attr(docsrs, doc(cfg(feature = "interpolate")))]
     fn interpolate(&self) -> Series;
@@ -426,6 +410,7 @@ pub trait SeriesTrait:
     }
 
     /// Unpack to ChunkedArray of dtype Time
+    #[cfg(feature = "dtype-time")]
     fn time(&self) -> Result<&TimeChunked> {
         Err(PolarsError::SchemaMisMatch(
             format!("Series dtype {:?} != Time", self.dtype()).into(),
@@ -433,6 +418,7 @@ pub trait SeriesTrait:
     }
 
     /// Unpack to ChunkedArray of dtype Date
+    #[cfg(feature = "dtype-date")]
     fn date(&self) -> Result<&DateChunked> {
         Err(PolarsError::SchemaMisMatch(
             format!(" Series dtype {:?} != Date", self.dtype()).into(),
@@ -440,6 +426,7 @@ pub trait SeriesTrait:
     }
 
     /// Unpack to ChunkedArray of dtype datetime
+    #[cfg(feature = "dtype-datetime")]
     fn datetime(&self) -> Result<&DatetimeChunked> {
         Err(PolarsError::SchemaMisMatch(
             format!("Series dtype {:?} != datetime", self.dtype()).into(),
@@ -447,6 +434,7 @@ pub trait SeriesTrait:
     }
 
     /// Unpack to ChunkedArray of dtype duration
+    #[cfg(feature = "dtype-duration")]
     fn duration(&self) -> Result<&DurationChunked> {
         Err(PolarsError::SchemaMisMatch(
             format!("Series dtype {:?} != duration", self.dtype()).into(),
@@ -461,6 +449,7 @@ pub trait SeriesTrait:
     }
 
     /// Unpack to ChunkedArray of dtype categorical
+    #[cfg(feature = "dtype-categorical")]
     fn categorical(&self) -> Result<&CategoricalChunked> {
         Err(PolarsError::SchemaMisMatch(
             format!("Series dtype {:?} != categorical", self.dtype()).into(),
@@ -485,11 +474,16 @@ pub trait SeriesTrait:
         invalid_operation_panic!(self)
     }
 
-    /// Append a Series of the same type in place.
+    #[doc(hidden)]
     fn append(&mut self, _other: &Series) -> Result<()> {
         invalid_operation_panic!(self)
     }
 
+    #[doc(hidden)]
+    fn extend(&mut self, _other: &Series) -> Result<()> {
+        invalid_operation_panic!(self)
+    }
+
     /// Filter by boolean mask. This operation clones data.
     fn filter(&self, _filter: &BooleanChunked) -> Result<Series> {
         invalid_operation_panic!(self)
@@ -514,7 +508,7 @@ pub trait SeriesTrait:
     ///
     /// # Safety
     /// This doesn't check any bounds.
-    unsafe fn take_unchecked(&self, _idx: &UInt32Chunked) -> Result<Series> {
+    unsafe fn take_unchecked(&self, _idx: &IdxCa) -> Result<Series> {
         invalid_operation_panic!(self)
     }
 
@@ -536,7 +530,7 @@ pub trait SeriesTrait:
     }
 
     /// Take by index. This operation is clone.
-    fn take(&self, _indices: &UInt32Chunked) -> Result<Series> {
+    fn take(&self, _indices: &IdxCa) -> Result<Series> {
         invalid_operation_panic!(self)
     }
 
@@ -602,10 +596,6 @@ pub trait SeriesTrait:
         invalid_operation_panic!(self)
     }
 
-    fn value_counts(&self) -> Result<DataFrame> {
-        invalid_operation_panic!(self)
-    }
-
     /// Get a single value by index. Don't use this operation for loops as a runtime cast is
     /// needed for every iteration.
     fn get(&self, _index: usize) -> AnyValue {
@@ -615,8 +605,11 @@ pub trait SeriesTrait:
     /// Get a single value by index. Don't use this operation for loops as a runtime cast is
     /// needed for every iteration.
     ///
+    /// This may refer to physical types
+    ///
     /// # Safety
     /// Does not do any bounds checking
+    #[cfg(feature = "private")]
     unsafe fn get_unchecked(&self, _index: usize) -> AnyValue {
         invalid_operation_panic!(self)
     }
@@ -626,7 +619,8 @@ pub trait SeriesTrait:
     }
 
     /// Retrieve the indexes needed for a sort.
-    fn argsort(&self, _reverse: bool) -> UInt32Chunked {
+    #[allow(unused)]
+    fn argsort(&self, options: SortOptions) -> IdxCa {
         invalid_operation_panic!(self)
     }
 
@@ -650,7 +644,7 @@ pub trait SeriesTrait:
     }
 
     /// Get first indexes of unique values.
-    fn arg_unique(&self) -> Result<UInt32Chunked> {
+    fn arg_unique(&self) -> Result<IdxCa> {
         invalid_operation_panic!(self)
     }
 
@@ -665,7 +659,7 @@ pub trait SeriesTrait:
     }
 
     /// Get indexes that evaluate true
-    fn arg_true(&self) -> Result<UInt32Chunked> {
+    fn arg_true(&self) -> Result<IdxCa> {
         Err(PolarsError::InvalidOperation(
             "arg_true can only be called for boolean dtype".into(),
         ))
@@ -789,10 +783,6 @@ pub trait SeriesTrait:
     fn min_as_series(&self) -> Series {
         invalid_operation_panic!(self)
     }
-    /// Get the mean of the Series as a new Series of length 1.
-    fn mean_as_series(&self) -> Series {
-        invalid_operation_panic!(self)
-    }
     /// Get the median of the Series as a new Series of length 1.
     fn median_as_series(&self) -> Series {
         invalid_operation_panic!(self)
@@ -998,28 +988,16 @@ pub trait SeriesTrait:
         }
     }
 
-    #[cfg(feature = "temporal")]
+    #[cfg(all(
+        feature = "temporal",
+        feature = "dtype-date",
+        feature = "dtype-datetime"
+    ))]
     #[cfg_attr(docsrs, doc(cfg(feature = "temporal")))]
-    /// Convert date(time) object to timestamp in ms.
-    fn timestamp(&self) -> Result<Int64Chunked> {
-        match self.dtype() {
-            DataType::Date => self
-                .cast(&DataType::Datetime(TimeUnit::Milliseconds, None))
-                .unwrap()
-                .datetime()
-                .map(|ca| (ca.deref().clone())),
-            DataType::Datetime(tu, tz) => {
-                use TimeUnit::*;
-                match (tu, tz) {
-                    (Nanoseconds, None) => self.datetime().map(|ca| ca.deref().clone() / 1_000_000),
-                    (Milliseconds, None) => self.datetime().map(|ca| ca.deref().clone()),
-                    _ => todo!(),
-                }
-            }
-            _ => Err(PolarsError::InvalidOperation(
-                format!("operation not supported on dtype {:?}", self.dtype()).into(),
-            )),
-        }
+    /// Convert date(time) object to timestamp in [`TimeUnit`].
+    fn timestamp(&self, tu: TimeUnit) -> Result<Int64Chunked> {
+        self.cast(&DataType::Datetime(tu, None))
+            .map(|s| s.datetime().unwrap().deref().clone())
     }
 
     /// Clone inner ChunkedArray and wrap in a new Arc
@@ -1036,8 +1014,6 @@ pub trait SeriesTrait:
 
     /// Get a hold to self as `Any` trait reference.
     /// Only implemented for ObjectType
-    #[cfg(feature = "object")]
-    #[cfg_attr(docsrs, doc(cfg(feature = "object")))]
     fn as_any(&self) -> &dyn Any {
         invalid_operation_panic!(self)
     }
@@ -1067,7 +1043,7 @@ pub trait SeriesTrait:
     }
     #[cfg(feature = "repeat_by")]
     #[cfg_attr(docsrs, doc(cfg(feature = "repeat_by")))]
-    fn repeat_by(&self, _by: &UInt32Chunked) -> ListChunked {
+    fn repeat_by(&self, _by: &IdxCa) -> ListChunked {
         invalid_operation_panic!(self)
     }
     #[cfg(feature = "checked_arithmetic")]
