@@ -3,8 +3,6 @@ pub(crate) mod series;
 use crate::prelude::*;
 use crate::POOL;
 pub use arrow;
-#[cfg(feature = "temporal")]
-pub use chrono;
 pub use num_cpus;
 pub use polars_arrow::utils::TrustMyLength;
 pub use polars_arrow::utils::*;
@@ -13,6 +11,9 @@ use rayon::prelude::*;
 use std::borrow::Cow;
 use std::ops::{Deref, DerefMut};
 
+#[cfg(feature = "private")]
+pub use crate::chunked_array::ops::sort::argsort_no_nulls;
+
 #[repr(transparent)]
 pub struct Wrap<T>(pub T);
 
@@ -23,22 +24,6 @@ impl<T> Deref for Wrap<T> {
     }
 }
 
-unsafe fn index_of_unchecked<T>(slice: &[T], item: &T) -> usize {
-    (item as *const _ as usize - slice.as_ptr() as usize) / std::mem::size_of::<T>()
-}
-
-fn index_of<T>(slice: &[T], item: &T) -> Option<usize> {
-    debug_assert!(std::mem::size_of::<T>() > 0);
-    let ptr = item as *const T;
-    unsafe {
-        if slice.as_ptr() < ptr && slice.as_ptr().add(slice.len()) > ptr {
-            Some(index_of_unchecked(slice, item))
-        } else {
-            None
-        }
-    }
-}
-
 pub(crate) fn set_partition_size() -> usize {
     let mut n_partitions = POOL.current_num_threads();
     // set n_partitions to closes 2^n above the no of threads.
@@ -117,10 +102,34 @@ macro_rules! split_array {
 }
 
 #[cfg(feature = "private")]
-pub fn split_ca<T>(ca: &ChunkedArray<T>, n: usize) -> Result<Vec<ChunkedArray<T>>> {
+pub fn split_ca<T>(ca: &ChunkedArray<T>, n: usize) -> Result<Vec<ChunkedArray<T>>>
+where
+    ChunkedArray<T>: ChunkOps,
+{
     split_array!(ca, n, i64)
 }
 
+// prefer this one over split_ca, as this can push the null_count into the thread pool
+pub fn split_offsets(len: usize, n: usize) -> Vec<(usize, usize)> {
+    if n == 1 {
+        vec![(0, len)]
+    } else {
+        let chunk_size = len / n;
+
+        (0..n)
+            .map(|partition| {
+                let offset = partition * chunk_size;
+                let len = if partition == (n - 1) {
+                    len - offset
+                } else {
+                    chunk_size
+                };
+                (partition * chunk_size, len)
+            })
+            .collect_trusted()
+    }
+}
+
 #[cfg(feature = "private")]
 pub fn split_series(s: &Series, n: usize) -> Result<Vec<Series>> {
     split_array!(s, n, i64)
@@ -162,105 +171,6 @@ pub fn slice_offsets(offset: i64, length: usize, array_len: usize) -> (usize, us
     }
 }
 
-#[derive(Clone, Copy, Debug, PartialEq, Eq, Hash)]
-pub struct Node(pub usize);
-
-impl Default for Node {
-    fn default() -> Self {
-        Node(usize::MAX)
-    }
-}
-
-#[derive(Clone)]
-#[cfg(feature = "private")]
-pub struct Arena<T> {
-    items: Vec<T>,
-}
-
-impl<T> Default for Arena<T> {
-    fn default() -> Self {
-        Self::new()
-    }
-}
-
-/// Simple Arena implementation
-/// Allocates memory and stores item in a Vec. Only deallocates when being dropped itself.
-impl<T> Arena<T> {
-    pub fn add(&mut self, val: T) -> Node {
-        let idx = self.items.len();
-        self.items.push(val);
-        Node(idx)
-    }
-
-    pub fn pop(&mut self) -> Option<T> {
-        self.items.pop()
-    }
-
-    pub fn len(&self) -> usize {
-        self.items.len()
-    }
-
-    pub fn is_empty(&self) -> bool {
-        self.items.is_empty()
-    }
-
-    pub fn new() -> Self {
-        Arena { items: vec![] }
-    }
-
-    pub fn with_capacity(cap: usize) -> Self {
-        Arena {
-            items: Vec::with_capacity(cap),
-        }
-    }
-
-    pub fn get_node(&self, val: &T) -> Option<Node> {
-        index_of(&self.items, val).map(Node)
-    }
-
-    #[inline]
-    pub fn get(&self, idx: Node) -> &T {
-        debug_assert!(idx.0 < self.items.len());
-        unsafe { self.items.get_unchecked(idx.0) }
-    }
-
-    #[inline]
-    pub fn get_mut(&mut self, idx: Node) -> &mut T {
-        debug_assert!(idx.0 < self.items.len());
-        unsafe { self.items.get_unchecked_mut(idx.0) }
-    }
-
-    #[inline]
-    pub fn replace(&mut self, idx: Node, val: T) {
-        let x = self.get_mut(idx);
-        *x = val;
-    }
-}
-
-impl<T: Default> Arena<T> {
-    #[inline]
-    pub fn take(&mut self, idx: Node) -> T {
-        std::mem::take(self.get_mut(idx))
-    }
-
-    pub fn replace_with<F>(&mut self, idx: Node, f: F)
-    where
-        F: FnOnce(T) -> T,
-    {
-        let val = self.take(idx);
-        self.replace(idx, f(val));
-    }
-
-    pub fn try_replace_with<F>(&mut self, idx: Node, mut f: F) -> Result<()>
-    where
-        F: FnMut(T) -> Result<T>,
-    {
-        let val = self.take(idx);
-        self.replace(idx, f(val)?);
-        Ok(())
-    }
-}
-
 /// Apply a macro on the Series
 #[macro_export]
 macro_rules! match_dtype_to_physical_apply_macro {
@@ -385,7 +295,7 @@ macro_rules! apply_method_all_arrow_series {
             DataType::Float32 => $self.f32().unwrap().$method($($args),*),
             DataType::Float64 => $self.f64().unwrap().$method($($args),*),
             DataType::Date => $self.date().unwrap().$method($($args),*),
-            DataType::Datetime=> $self.datetime().unwrap().$method($($args),*),
+            DataType::Datetime(_, _) => $self.datetime().unwrap().$method($($args),*),
             DataType::List(_) => $self.list().unwrap().$method($($args),*),
             dt => panic!("dtype {:?} not supported", dt)
         }
@@ -415,7 +325,7 @@ macro_rules! apply_method_numeric_series {
             #[cfg(feature = "dtype-date")]
             DataType::Date => $self.date().unwrap().$method($($args),*),
             #[cfg(feature = "dtype-datetime")]
-            DataType::Datetime=> $self.datetime().unwrap().$method($($args),*),
+            DataType::Datetime(_, _) => $self.datetime().unwrap().$method($($args),*),
             _ => unimplemented!(),
         }
     }
@@ -470,6 +380,16 @@ macro_rules! df {
     }
 }
 
+#[cfg(feature = "private")]
+pub fn get_time_units(tu_l: &TimeUnit, tu_r: &TimeUnit) -> TimeUnit {
+    use TimeUnit::*;
+    match (tu_l, tu_r) {
+        (Nanoseconds, Microseconds) => Microseconds,
+        (_, Milliseconds) => Milliseconds,
+        _ => *tu_l,
+    }
+}
+
 /// Given two datatypes, determine the supertype that both types can safely be cast to
 #[cfg(feature = "private")]
 pub fn get_supertype(l: &DataType, r: &DataType) -> Result<DataType> {
@@ -540,6 +460,12 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         (UInt32, Float32) => Some(Float32),
         (UInt32, Float64) => Some(Float64),
         (UInt32, Boolean) => Some(UInt32),
+        #[cfg(feature = "dtype-date")]
+        (UInt32, Date) => Some(Int64),
+        #[cfg(feature = "dtype-datetime")]
+        (UInt32, Datetime(_, _)) => Some(Int64),
+        #[cfg(feature = "dtype-duration")]
+        (UInt32, Duration(_)) => Some(Int64),
 
         (UInt64, UInt8) => Some(UInt64),
         (UInt64, UInt16) => Some(UInt64),
@@ -574,7 +500,9 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         #[cfg(feature = "dtype-date")]
         (Int32, Date) => Some(Int32),
         #[cfg(feature = "dtype-datetime")]
-        (Int32, Datetime) => Some(Int64),
+        (Int32, Datetime(_, _)) => Some(Int64),
+        #[cfg(feature = "dtype-duration")]
+        (Int32, Duration(_)) => Some(Int64),
         #[cfg(feature = "dtype-time")]
         (Int32, Time) => Some(Int64),
         (Int32, Boolean) => Some(Int32),
@@ -586,7 +514,9 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         (Int64, Float32) => Some(Float32),
         (Int64, Float64) => Some(Float64),
         #[cfg(feature = "dtype-datetime")]
-        (Int64, Datetime) => Some(Int64),
+        (Int64, Datetime(_, _)) => Some(Int64),
+        #[cfg(feature = "dtype-duration")]
+        (Int64, Duration(_)) => Some(Int64),
         #[cfg(feature = "dtype-date")]
         (Int64, Date) => Some(Int32),
         #[cfg(feature = "dtype-time")]
@@ -598,7 +528,9 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         #[cfg(feature = "dtype-date")]
         (Float32, Date) => Some(Float32),
         #[cfg(feature = "dtype-datetime")]
-        (Float32, Datetime) => Some(Float64),
+        (Float32, Datetime(_, _)) => Some(Float64),
+        #[cfg(feature = "dtype-duration")]
+        (Float32, Duration(_)) => Some(Float64),
         #[cfg(feature = "dtype-time")]
         (Float32, Time) => Some(Float64),
         (Float64, Float32) => Some(Float64),
@@ -606,11 +538,17 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         #[cfg(feature = "dtype-date")]
         (Float64, Date) => Some(Float64),
         #[cfg(feature = "dtype-datetime")]
-        (Float64, Datetime) => Some(Float64),
+        (Float64, Datetime(_, _)) => Some(Float64),
+        #[cfg(feature = "dtype-duration")]
+        (Float64, Duration(_)) => Some(Float64),
         #[cfg(feature = "dtype-time")]
         (Float64, Time) => Some(Float64),
         (Float64, Boolean) => Some(Float64),
 
+        #[cfg(feature = "dtype-datetime")]
+        (Date, UInt32) => Some(Int64),
+        #[cfg(feature = "dtype-datetime")]
+        (Date, UInt64) => Some(Int64),
         #[cfg(feature = "dtype-datetime")]
         (Date, Int32) => Some(Int32),
         #[cfg(feature = "dtype-datetime")]
@@ -620,18 +558,35 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         #[cfg(feature = "dtype-datetime")]
         (Date, Float64) => Some(Float64),
         #[cfg(feature = "dtype-datetime")]
-        (Date, Datetime) => Some(Datetime),
+        (Date, Datetime(tu, tz)) => Some(Datetime(*tu, tz.clone())),
 
         #[cfg(feature = "dtype-date")]
-        (Datetime, Int32) => Some(Int64),
+        (Datetime(_, _), UInt32) => Some(Int64),
         #[cfg(feature = "dtype-date")]
-        (Datetime, Int64) => Some(Int64),
+        (Datetime(_, _), UInt64) => Some(Int64),
         #[cfg(feature = "dtype-date")]
-        (Datetime, Float32) => Some(Float64),
+        (Datetime(_, _), Int32) => Some(Int64),
         #[cfg(feature = "dtype-date")]
-        (Datetime, Float64) => Some(Float64),
+        (Datetime(_, _), Int64) => Some(Int64),
         #[cfg(feature = "dtype-date")]
-        (Datetime, Date) => Some(Datetime),
+        (Datetime(_, _), Float32) => Some(Float64),
+        #[cfg(feature = "dtype-date")]
+        (Datetime(_, _), Float64) => Some(Float64),
+        #[cfg(feature = "dtype-date")]
+        (Datetime(tu, tz), Date) => Some(Datetime(*tu, tz.clone())),
+
+        #[cfg(feature = "dtype-duration")]
+        (Duration(_), UInt32) => Some(Int64),
+        #[cfg(feature = "dtype-duration")]
+        (Duration(_), UInt64) => Some(Int64),
+        #[cfg(feature = "dtype-duration")]
+        (Duration(_), Int32) => Some(Int64),
+        #[cfg(feature = "dtype-duration")]
+        (Duration(_), Int64) => Some(Int64),
+        #[cfg(feature = "dtype-duration")]
+        (Duration(_), Float32) => Some(Float64),
+        #[cfg(feature = "dtype-duration")]
+        (Duration(_), Float64) => Some(Float64),
 
         #[cfg(feature = "dtype-time")]
         (Time, Int32) => Some(Int64),
@@ -643,9 +598,9 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         (Time, Float64) => Some(Float64),
 
         #[cfg(all(feature = "dtype-time", feature = "dtype-datetime"))]
-        (Time, Datetime) => Some(Int64),
+        (Time, Datetime(_, _)) => Some(Int64),
         #[cfg(all(feature = "dtype-time", feature = "dtype-datetime"))]
-        (Datetime, Time) => Some(Int64),
+        (Datetime(_, _), Time) => Some(Int64),
         #[cfg(all(feature = "dtype-time", feature = "dtype-date"))]
         (Time, Date) => Some(Int64),
         #[cfg(all(feature = "dtype-time", feature = "dtype-date"))]
@@ -669,10 +624,46 @@ fn _get_supertype(l: &DataType, r: &DataType) -> Option<DataType> {
         (dt, Null) => Some(dt.clone()),
         (Null, dt) => Some(dt.clone()),
 
+        (Duration(lu), Datetime(ru, Some(tz))) | (Datetime(lu, Some(tz)), Duration(ru)) => {
+            if tz.is_empty() {
+                Some(Datetime(get_time_units(lu, ru), None))
+            } else {
+                Some(Datetime(get_time_units(lu, ru), Some(tz.clone())))
+            }
+        }
+        (Duration(lu), Datetime(ru, None)) | (Datetime(lu, None), Duration(ru)) => {
+            Some(Datetime(get_time_units(lu, ru), None))
+        }
+        (Duration(_), Date) | (Date, Duration(_)) => Some(Datetime(TimeUnit::Milliseconds, None)),
+        (Duration(lu), Duration(ru)) => Some(Duration(get_time_units(lu, ru))),
+
+        // None and Some("") timezones
+        // we cast from more precision to higher precision as that always fits with occasional loss of precision
+        (Datetime(tu_l, tz_l), Datetime(tu_r, tz_r))
+            if (tz_l.is_none() || tz_l.as_deref() == Some(""))
+                && (tz_r.is_none() || tz_r.as_deref() == Some("")) =>
+        {
+            let tu = get_time_units(tu_l, tu_r);
+            Some(Datetime(tu, None))
+        }
         _ => None,
     }
 }
 
+/// This takes ownership of the DataFrame so that drop is called earlier.
+/// Does not check if schema is correct
+pub fn accumulate_dataframes_vertical_unchecked<I>(dfs: I) -> Result<DataFrame>
+where
+    I: IntoIterator<Item = DataFrame>,
+{
+    let mut iter = dfs.into_iter();
+    let mut acc_df = iter.next().unwrap();
+    for df in iter {
+        acc_df.vstack_mut_unchecked(&df);
+    }
+    Ok(acc_df)
+}
+
 /// This takes ownership of the DataFrame so that drop is called earlier.
 pub fn accumulate_dataframes_vertical<I>(dfs: I) -> Result<DataFrame>
 where
@@ -716,9 +707,17 @@ where
     F: Fn(Series) -> Result<Series> + Send + Sync,
 {
     let n_threads = n_threads.unwrap_or_else(|| POOL.current_num_threads());
-    let slices = split_series(&s, n_threads)?;
+    let splits = split_offsets(s.len(), n_threads);
 
-    let chunks = POOL.install(|| slices.into_par_iter().map(&f).collect::<Result<Vec<_>>>())?;
+    let chunks = POOL.install(|| {
+        splits
+            .into_par_iter()
+            .map(|(offset, len)| {
+                let s = s.slice(offset as i64, len);
+                f(s)
+            })
+            .collect::<Result<Vec<_>>>()
+    })?;
 
     let mut iter = chunks.into_iter();
     let first = iter.next().unwrap();
@@ -837,18 +836,31 @@ pub trait IntoVec<T> {
     fn into_vec(self) -> Vec<T>;
 }
 
+pub trait Arg {}
+impl Arg for bool {}
+
 impl IntoVec<bool> for bool {
     fn into_vec(self) -> Vec<bool> {
         vec![self]
     }
 }
 
-impl<T> IntoVec<T> for Vec<T> {
+impl<T: Arg> IntoVec<T> for Vec<T> {
     fn into_vec(self) -> Self {
         self
     }
 }
 
+impl<I, S> IntoVec<String> for I
+where
+    I: IntoIterator<Item = S>,
+    S: AsRef<str>,
+{
+    fn into_vec(self) -> Vec<String> {
+        self.into_iter().map(|s| s.as_ref().to_string()).collect()
+    }
+}
+
 /// This logic is same as the impl on ChunkedArray
 /// The difference is that there is less indirection because the caller should preallocate
 /// `chunk_lens` once. On the `ChunkedArray` we indirect through an `ArrayRef` which is an indirection
@@ -888,9 +900,9 @@ mod test {
 
     #[test]
     fn test_align_chunks() {
-        let a = Int32Chunked::new_from_slice("", &[1, 2, 3, 4]);
-        let mut b = Int32Chunked::new_from_slice("", &[1]);
-        let b2 = Int32Chunked::new_from_slice("", &[2, 3, 4]);
+        let a = Int32Chunked::new("", &[1, 2, 3, 4]);
+        let mut b = Int32Chunked::new("", &[1]);
+        let b2 = Int32Chunked::new("", &[2, 3, 4]);
 
         b.append(&b2);
         let (a, b) = align_chunks_binary(&a, &b);
@@ -899,8 +911,8 @@ mod test {
             b.chunk_id().collect::<Vec<_>>()
         );
 
-        let a = Int32Chunked::new_from_slice("", &[1, 2, 3, 4]);
-        let mut b = Int32Chunked::new_from_slice("", &[1]);
+        let a = Int32Chunked::new("", &[1, 2, 3, 4]);
+        let mut b = Int32Chunked::new("", &[1]);
         let b1 = b.clone();
         b.append(&b1);
         b.append(&b1);
